#!/usr/bin/env python3
"""Program's entry point."""
import sys
import time
from pathlib import Path

import humanize

from .argparse_wrapper import get_parsed_args
from .logs import get_logger


def main():
    """Program's main routine."""
    t_start = time.time()
    prog_name = Path(sys.argv[0]).name
    args = get_parsed_args(prog_name)
    logger = get_logger(prog_name, args.loglevel)
    if args.command is None:
        # Enforce this here instead of on the command's parser.add_subparsers
        # definition because add_subparsers only introduced the "required" arg
        # from python v3.7, and we want the code to work with python>=3.6.10.
        msg = "Cannot run {0} without a command! Please run "
        msg += "'{0} -h' for help."
        msg = msg.format(prog_name)
        logger.error(msg)
        sys.exit(1)
    args.func(args)
    elapsed = time.time() - t_start
    if elapsed >= 60:
        logger.debug(
            "Leaving 'main()'. Elapsed: %.2fs (~%s)",
            elapsed,
            humanize.precisedelta(elapsed),
        )
    else:
        logger.debug("Leaving 'main()'. Elapsed: %.2fs", elapsed)



#!/usr/bin/env python3
"""Code related to outlier removal functionality."""
import logging
import time

import numpy as np
from numba import njit, prange
from sklearn.neighbors import LocalOutlierFactor

logger = logging.getLogger(__name__)


# Routines related to the "iterative" outlier removal method
@njit(cache=True)
def _trim_np_array(array, perc, min_size=3):
    """Remove extreme values from both sides of an array.

    Takes an array and removes the "50*perc %" largest and
    "50*perc %" smallest elements, provided that the resulting
    trimmed array has a length no smaller tyan min_size

    Args:
        array (numpy.array): Input array.
        perc (float): Proportion of the array elements to be removed. Should be
            between 0 and 1.
        min_size (int): Size below which the array does not get truncated.
            (Default value = 3)

    Returns:
        numpy.array: Truncated array.

    """
    len_array = len(array)
    n_rm_each_side = int(abs(perc) * 0.5 * len_array)
    truncated_size = len_array - 2 * n_rm_each_side
    if (truncated_size != len_array) and (truncated_size >= min_size):
        array.sort()
        array = array[n_rm_each_side:-n_rm_each_side]
    return array


@njit(cache=True, parallel=True)
def _truncated_means_and_stds(df, perc, min_size=3):
    """Return truncated means and stdevs for all columns in the dataframe df.

    Takes a numpy version "df" of a pandas dataframe (obtained by
    calling the ".to_numpy" method on the original dataframe) and

    Args:
        df (numpy.ndarray): The input data.
        perc (float): Proportion of the array elements to be removedi for the
            truncation used in the calculation of the stds and means. Should be
            between 0 and 1.
        min_size (int): Size below which the data column arrays don't get
            truncated in the calculation of the stds and means.
            (Default value = 3)

    Returns:
        {
            numpy.array: Truncated means calculated for each column in the
                input data.
            numpy.array: Truncated stds for each column in the data.
        }

    """
    ncols = df.shape[1]
    means = np.zeros(ncols)
    stds = np.zeros(ncols)
    for icol in prange(ncols):
        array = _trim_np_array(df[:, icol], perc, min_size)
        means[icol] = np.mean(array)
        stds[icol] = np.std(array)
    return means, stds


@njit(cache=True, parallel=True)
def _filter_outliers_iterative_one_iter(
    df, max_n_stdev_around_mean, truncate, weights
):
    """Execute one iteration of the _filter_outliers_iterative function.

    Args:
        df (numpy.ndarray): Matrix (obtained from a dataframe) containing
            clustering info. The cluster labels are assumed to have been
            stored in the last column of df.
        max_n_stdev_around_mean (float): Max allowed number of stdevs around
            the mean for an observation to be considered acceptable.
        truncate (float): Proportion of array elements to be removed for the
            calculation of the truncated stds and means. Should lie between
            0 and 1.
        weights (numpy.array): Weights chosen for each observation parameter.

    Returns:
        numpy.ndarray: Input matrix with cluster labels for outlier obs
            reassigned to -2.

    """
    unique_labels = np.unique(df[:, -1])
    for label in unique_labels[unique_labels > -1]:
        indices_for_label = np.transpose(np.argwhere(df[:, -1] == label))[0]

        cols_taken_into_acc = []
        for icol in range(df.shape[1] - 1):
            if icol in [0, 1]:
                if weights[0] < 1e-3:
                    continue
            elif weights[icol - 1] < 1e-3:
                continue
            cols_taken_into_acc.append(icol)
        cols_taken_into_acc = np.array(cols_taken_into_acc)

        df_subset = df[indices_for_label, :][:, cols_taken_into_acc]
        means, stds = _truncated_means_and_stds(df_subset, truncate)
        tol = max_n_stdev_around_mean * stds[:]
        for irow in indices_for_label:
            for icol in cols_taken_into_acc:
                if np.abs(df[irow, icol] - means[icol]) > tol[icol]:
                    # Use "-2" as a "removed by refining methods" flag
                    df[irow, -1] = -2
                    break

    return df


def _filter_outliers_iterative(
    df,
    max_num_iter=100,
    max_n_stdev_around_mean=2.0,
    trunc_perc=0.0,
    weights=None,
):
    """Detect & assign new labels to obs where abs(obs[i]-obs_mean)>tolerance.

    Helper function for the filter_outliers_iterative routine.

    Args:
        df (pandas.Dataframe): Dataframe containing clustering info.
        max_num_iter (int): Max number of iterations performed.
            (Default value = 100).
        max_n_stdev_around_mean (float): Max allowed number of stdevs around
            the mean for an observation to be considered acceptable.
            (Default value = 2.0).
        trunc_perc (float): Proportion of array elements to be removed for the
            calculation of the truncated stds and means. Should lie between
            0 and 1. (Default value = 0.0).
        weights (numpy.array): Weights chosen for each observation parameter.
            If None is passed, then all obs will have the same weight.
            (Default value = None).

    Returns:
        numpy.array: Array with clustering labels for each row in the input
            data. Entries considered to be outliers by this routine will be
            assigned the integer cluster label "-2".

    """
    used_cols = [
        "lat",
        "lon",
        "alt",
        "mslp",
        "pressure",
        "temperature",
        "humidity",
        "sum_rain_1",
        "cluster_label",
    ]
    df = df.loc[:, df.columns.intersection(used_cols)].to_numpy()
    if weights is None:
        # (lat, lon) -> geodist weight; remove "cluster_label" from weights
        weights = np.ones(
            len([_ for _ in df.columns if _ not in ["lat", "cluster_label"]])
        )
    # Set any negative weight value to zero
    weights = np.where(weights < 0, 0.0, weights)

    n_removed_tot = 0
    verbose = (logger.getEffectiveLevel() == logging.DEBUG,)
    for i in range(max_num_iter):
        if verbose:
            print("    > Refining scan, iteration ", i + 1, ": ", end=" ")
            start_time = time.time()
        # We use "-2" as a "removed by refining methods" flag
        n_removed_old = np.count_nonzero(df[:, -1] == -2)
        df = _filter_outliers_iterative_one_iter(
            df, max_n_stdev_around_mean, truncate=trunc_perc, weights=weights,
        )
        n_removed_new = np.count_nonzero(df[:, -1] == -2)
        n_removed_this_iter = n_removed_new - n_removed_old
        n_removed_tot = n_removed_tot + n_removed_this_iter
        if verbose:
            print("* rm ", n_removed_this_iter, end=" ")
            print(" stations, ", n_removed_tot, " so far. ", end=" ")
            print("Took ", time.time() - start_time, "s")
        if n_removed_this_iter == 0:
            break
    return df[:, -1]


def filter_outliers_iterative(
    df,
    skip,
    weights,
    trunc_perc=0.25,
    max_num_refine_iter=1000,
    max_n_stdev_around_mean=2.0,
):
    """Filter outliers using an iterative method.

    For each obs type (column) in the dataframe, calculate the mean value
    and standard deviation. Remove rows for which any observation has a value
    lying more than "max_n_stdev_around_mean" stdevs from the corresponding
    column's mean. Repeat the process untill either no data is removed or
    the max number of iterations is reached.

    Args:
        df (pandas.Dataframe): Dataframe containing clustering info.
        skip (list): List of names of columns in the df to be skipped (i.e.,
            not taken into account in the filtering).
        weights (numpy.array): Weights chosen for each observation parameter.
            If None is passed, then all obs will have the same weight.
        trunc_perc (float): Proportion of array elements to be removed for the
            calculation of the truncated stds and means. Should lie between
            0 and 1. (Default value = 0.25).
        max_num_refine_iter (int): Max number of iterations performed.
            (Default value = 1000).
        max_n_stdev_around_mean (float): Max allowed number of stdevs around
            the mean for an observation to be considered acceptable.
            (Default value = 2.0).

    Returns:
        pandas.Dataframe: Copy of input data, with clustering labels reassigned
            such that outliers are marked with integer label "-2".

    """
    df["cluster_label"] = _filter_outliers_iterative(
        df.drop(list(skip), axis=1),
        max_num_iter=max_num_refine_iter,
        max_n_stdev_around_mean=max_n_stdev_around_mean,
        trunc_perc=trunc_perc,
        weights=weights,
    ).astype(int)
    return df


# GLOSH outlier removal method
def filter_outliers_glosh(df, outlier_scores):
    """Detect outliers using the GLOSH method.

    Args:
        df (pandas.Dataframe): Input data with clustering info.
        outlier_scores (numpy.array): GLOSH outlier scores for each row
            in the input dataframe.

    Returns:
        pandas.Dataframe: Copy of input data, with clustering labels reassigned
            such that outliers are marked with integer label "-2".

    """
    df["GLOSH"] = outlier_scores
    threshold = min(0.25, df[df["cluster_label"] > -1]["GLOSH"].quantile(0.75))
    outliers_index = (df["cluster_label"] > -1) & (df["GLOSH"] > threshold)
    # Use "-2" as a "removed by refining methods" flag
    df.at[outliers_index, "cluster_label"] = -2
    return df


# Routines related to the LOF outlier removal method
def get_local_outlier_factors(df, distance_matrix, calc_per_cluster=False):
    """Calculate local outlier factors (LOF) for entries in the input data.

    Args:
        df (pandas.Dataframe): Input data with clustering info.
        distance_matrix (netatmoqc.metrics.HollowSymmetricMatrix): Distance
            matrix consistent with the input data.
        calc_per_cluster (bool): Calculate the LOF scores in a per-cluster
            basis (True) or using the dataset as a whole (False).
            (Default value = False).

    Returns:
        numpy.array: Local outlier factors (LOF) for every row of the
            input data.

    """
    # See <https://scikit-learn.org/stable/modules/generated/
    # sklearn.neighbors.LocalOutlierFactor.html#>
    unique_labels = df["cluster_label"].unique()
    all_lof_values = np.empty(len(df.index))
    all_lof_values[:] = np.nan
    # Is it better to do this in a per-cluster basis or
    # with the dataset as a whole? Maybe, with such a small
    # n_neighbors, the results won't change much.
    if calc_per_cluster:
        for label in unique_labels:
            if label < 0:
                continue
            indices = df.index[df["cluster_label"] == label]
            clf = LocalOutlierFactor(
                n_neighbors=min(3, len(indices)), metric="precomputed"
            )
            clf.fit_predict(distance_matrix.subspace(indices))
            all_lof_values[indices] = clf.negative_outlier_factor_
    else:
        indices = df.index[df["cluster_label"] > -1]
        clf = LocalOutlierFactor(n_neighbors=3, metric="precomputed")
        clf.fit_predict(distance_matrix.subspace(indices))
        all_lof_values[indices] = clf.negative_outlier_factor_
    return all_lof_values


def filter_outliers_lof(df, distance_matrix):
    """Filter outliers based on LOF scores.

    Args:
        df (pandas.Dataframe): Input data with clustering info.
        distance_matrix (netatmoqc.metrics.HollowSymmetricMatrix): Distance
            matrix consistent with the input data.

    Returns:
        pandas.Dataframe: Copy of input data, with clustering labels for
        outlier observations reassigned with integer label "-2".

    """
    df["LOF"] = get_local_outlier_factors(df, distance_matrix)
    outliers_index = (df["cluster_label"] > -1) & (df["LOF"] < -1.5)
    # Use "-2" as a "removed by refining methods" flag
    df.at[outliers_index, "cluster_label"] = -2
    return df


# Higher-level outlier removal routine calling the specific ones defined above
def filter_outliers(df, db, outlier_rm_method, distance_matrix, **kwargs):
    """Filter outliers according to specified outlier detection method.

    Args:
        df (pandas.Dataframe): Input data with clustering info.
        db (obj): Output of clustering method.
        outlier_rm_method (str): Outlier detection method of choice.
        distance_matrix (netatmoqc.metrics.HollowSymmetricMatrix): Distance
            matrix consistent with the input data.
        **kwargs: Passed on to internalm wrapped routines.

    Returns:
        pandas.Dataframe: Copy of input data, with clustering labels for
        outlier observations reassigned with integer label "-2".

    """
    tstart = time.time()
    logger.debug(
        '    > Running outlier removal method "%s" with kwargs=%s',
        outlier_rm_method,
        kwargs,
    )
    # Copy original cluster labels to a new column, so we can keep track
    # of where the removed obs came from
    df["original_cluster_label"] = df["cluster_label"]
    if outlier_rm_method == "glosh":
        rtn = filter_outliers_glosh(df, outlier_scores=db.outlier_scores_)
    elif outlier_rm_method == "lof":
        rtn = filter_outliers_lof(df, distance_matrix=distance_matrix)
    elif outlier_rm_method == "reclustering":
        func = kwargs.pop("reclustering_function")
        rtn = func(df, distance_matrix, **kwargs)
    else:
        rtn = filter_outliers_iterative(df, **kwargs)
    logger.debug(
        "      * Done with outlier removal. Elapsed: %.1fs",
        time.time() - tstart,
    )

    return rtn



#!/usr/bin/env python3
"""Implement HollowSymmetricMatrix class and related routines."""
import copy
import logging
import time

import humanize
import numpy as np
from numba import njit, prange

logger = logging.getLogger(__name__)


class HsmDiagNotStoredInCompactForm(Exception):
    """Exception to inform that HSM matrix is not in compact form."""

    def __init__(
        self,
        *args,
        msg="Diagonal elements of HSM matrix are not stored in compact form",
        **kwargs
    ):
        """Initialise with a default exception message."""
        super().__init__(msg, *args, **kwargs)


@njit(cache=True)
def _matrix_index_to_data_index(order, i, j):
    """Hollow symmetric matrix 'matrix -> compact array' index conversion.

    Given a hollow symmetric matrix M of order "order" and a linear array,
    say "M_data", that stores only the independent elements of M, return the
    index i_data such that

                M(i, j) = M_data(i_data)

    This function is used in the HollowSymmetricMatrix class.

    Args:
        order (int): Order of the hollow symmetric matrix.
        i (int): Matrix row index.
        j (int): Matrix column index.

    Returns:
        int:
            * In the interval [0, n*(n - 1)/2) if i != j
            * -(order + 1) if i == j (flagging thus that i==j elements are not
                stored).
                The choice of returning "-(order + 1)" for i==j instead of
                raising an exception is because numba only implemented support
                to handling exceptions in python < 3.7, but we need this to be
                used with python >= 3.6.10.

    Raises:
        ValueError: If order < 1.
        IndexError: If the condition "-order <= i,j < order" does not hold.

    """
    if order < 1:
        raise ValueError("Argument 'order' must be larger than zero")
    if i < -order or i >= order:
        # Numba doesn't allow a nicely formatted msg with indice values here
        raise IndexError("Index i should be larger than -(order + 1)")
    if j < -order or j >= order:
        raise IndexError("Index j should be larger than -(order + 1)")

    # Wrap around valid negative indices
    if i < 0:
        i += order
    if j < 0:
        j += order

    if i == j:
        # Flag that there's no index corresponding to i==j
        return -(order + 1)

    if i > j:
        # Enforce M(i, j) = M(j, i)
        i, j = j, i

    # n_indep_els_before_row: Total number of independent elements
    # located in the rows prior to row=i for the matrix M
    n_indep_els_before_row = i * order - (i * (i + 1)) // 2

    return n_indep_els_before_row + (j - i - 1)


@njit(cache=True)
def _data_index_to_matrix_index(n, i_data, check_bounds=True):
    """Hollow symmetric matrix 'compact array -> matrix' index conversion.

    Given a hollow symmetric matrix M of order "n" and a linear array, say
    "M_data", that stores only the independent elements of M, return the matrix
    indices (i, j) such that

                M(i, j) = M_data(i_data)

    This is the inverse of what is done in _matrix_index_to_data_index.

    This routine assumes
                        0 <= i_data < n(n-1)/2

    This function is used in the HollowSymmetricMatrix class, where
    consistency between the arguments is enforced.

    Row-major order.

    Args:
        n (int): Order of the hollow symmetric matrix (HSM).
        i_data (int): Index of data entry in the compact array that stores the
            independent elements of the HSM.
        check_bounds (bool): Whether or not to assert that n and i_data
            correspond to matrix indices (i, j) within bounds.
            (Default value = True).

    Returns:
        {
            int: Matrix row index,
            int: Matrix column index
        }

    Raises:
        ValueError: If check_bounds evaluates to True and n, i_data are outside
            their allowed ranges.

    """
    if check_bounds:
        if n < 1:
            raise ValueError("Arg 'n' must be larger than zero")
        if i_data < 0 or i_data >= n * (n - 1) / 2:
            raise ValueError("Arg 'i_data' not in the range [0, n*(n-1)/2)")

    i = (
        int(0.5 * ((2 * n + 1) - np.sqrt((2 * n + 1) ** 2 - 8 * (n + i_data))))
        - 1
    )
    j = (1 + i_data + i) - (i * (2 * n - i - 1)) // 2
    return i, j


@njit(cache=True, parallel=True)
def _to_dense(data, n, dtype):
    """Dense matrix form of a hollow symmetric matrix from its compact form.

    Takes an array 'data', assumed to hold the independent elements of a hollow
    symmetric M of order 'n', and returns a full (dense) representation of M.

    Assumes that len(data) = n*(n - 1)/2.
    No checks, however, are made on len(data). If you want to provide this
    functionality in an external API, please expose a wrapper to this function
    where this is asserted.

    This function is used in the HollowSymmetricMatrix class, where
    consistency between the arguments is enforced.

    Args:
        data (numpy.array): Array that stores the independent elements of the
            hollow symmetric matrix (hsm), i.e., the HSM's compact form.
        n (int): The order of the HSM.
        dtype (numpy.dtype): dtype of the data.

    Returns:
        np.nparray: Dense representation of the HSM (full 2D matrix form).

    """
    dense = np.zeros(shape=(n, n), dtype=dtype)
    for idata in prange(len(data)):
        i, j = _data_index_to_matrix_index(n, idata, check_bounds=False)
        data_value = data[idata]
        dense[i, j] = data_value
        dense[j, i] = data_value
    return dense


@njit(cache=True, parallel=True)
def _to_compact(dense, dtype):
    """Compact matrix form of a hollow symmetric matrix from its dense form.

    Args:
        dense (np.nparray): Dense representation of the HSM (full matrix form).
        dtype (numpy.dtype): dtype of the data.

    Returns:
        numpy.array: Array with the independent elements of the HSM

    Raises:
        ValueError: If dense is not a 2D, square matrix.

    """
    if dense.ndim != 2:
        raise ValueError("dense.ndim != 2")
    if dense.shape[0] != dense.shape[1]:
        raise ValueError("'dense' is not a square matrix")

    n = dense.shape[0]
    n_indep = (n ** 2 - n) // 2
    compact = np.zeros(n_indep, dtype=dtype)
    for idata in prange(len(compact)):
        i, j = _data_index_to_matrix_index(n, idata, check_bounds=False)
        compact[idata] = dense[i, j]
    return compact


@njit(cache=True)
def _get_hsm_subspace_from_compact(compact, indices):
    """Subspace of an HSM compact, as "irow in indices & icol=irow".

    Args:
        compact (numpy.array): Array with the independent elements of the HSM
        indices (list): Indices of the subdimentions of the HSM to include in
            the subspace.

    Returns:
        numpy.array: Compact form of the selected subspace.

    Raises:
        ValueError: If the number of elements in the subspace is not
            consistent with the passed indices.

    """
    n_indep = len(compact)
    n = int(np.rint(0.5 * (1.0 + np.sqrt(1.0 + 8.0 * n_indep))))
    new_n = len(indices)
    new_n_indep = (new_n ** 2 - new_n) // 2
    new_compact = np.zeros(new_n_indep, dtype=compact.dtype)

    idata_new = 0
    for i in indices:
        for j in indices:
            if i >= j:
                continue
            idata = _matrix_index_to_data_index(n, i, j)
            new_compact[idata_new] = compact[idata]
            idata_new += 1
    if idata_new != len(new_compact):
        raise ValueError("Number of calcd elements do not match the expected")
    return new_compact


@njit(cache=True)
def _get_submatrix_from_hsm_compact(hsm_compact, order, bcasted_keys):
    """Extract submatrix from a HollowSymmetricMatrix in dense form.

    Memory-efficient way to extract a submatrix from HollowSymmetricMatrix
    objects using fancy indexing. It does not require the conversion of the
    HollowSymmetricMatrix object into its dense form, and requires extra mem
    allocation for the submatrix elemnts only.

    Used in the __getitem__ method of HollowSymmetricMatrix

    Args:
        hsm_compact (numpy.array): Array with the compact form of the HSM.
        order (int): Order of the HSM.
        bcasted_keys (list): Broadcasted keys passed to the __getitem__
            method that calls this function. See also the function
            _getitem_keys_to_bcast_arrays.

    Returns:
        np.ndarray: Submatrix selected according to the indices contructed
            from bcasted_keys.

    """
    submatrix_shape = bcasted_keys[0].shape
    n_data = bcasted_keys[0].size
    submatrix_as_linear = np.empty(n_data, dtype=hsm_compact.dtype)
    i_data = 0
    for i_list, j_list in zip(bcasted_keys[0], bcasted_keys[1]):
        for i, j in zip(i_list, j_list):
            i_data_in_compact = _matrix_index_to_data_index(order, i, j)
            if i_data_in_compact < -order:
                # If _matrix_index_to_data_index flags "value not stored"
                submatrix_as_linear[i_data] = 0.0
            else:
                submatrix_as_linear[i_data] = hsm_compact[i_data_in_compact]
            i_data += 1
    return submatrix_as_linear.reshape(submatrix_shape)


def _getitem_keys_to_bcast_arrays(keys):
    """Convert keys passed to __getitem__ into broadcasted arrays."""
    if any(isinstance(key, slice) for key in keys):
        raise ValueError("Cannot accept slices as input")
    # Numba likes it better if we return an ndarray here
    return np.array(np.broadcast_arrays(*keys))


@njit(cache=True)
def _slice2range(the_slice, array_length):
    """Convert a slice into a list of indices according to the array length."""
    return range(*the_slice.indices(array_length))


@njit(cache=True)
def _get_hsm_double_slice(hsm_compact, order, row_slice, col_slice):
    """Extract [row_slice, col_slice] from HSM compact form."""
    row_numbers = _slice2range(row_slice, order)
    col_numbers = _slice2range(col_slice, order)
    submatrix_shape = (len(row_numbers), len(col_numbers))
    ndata = len(row_numbers) * len(col_numbers)
    submatrix = np.empty(ndata, dtype=hsm_compact.dtype)
    i_data = 0
    for i in row_numbers:
        for j in col_numbers:
            i_data_in_compact = _matrix_index_to_data_index(order, i, j)
            if i_data_in_compact < -order:
                # _matrix_index_to_data_index flag for "value is not stored"
                submatrix[i_data] = 0.0
            else:
                submatrix[i_data] = hsm_compact[i_data_in_compact]
            i_data += 1
    return submatrix.reshape(submatrix_shape)


def _get_hsm_row_slice(hsm_compact, order, the_slice):
    """Extract row slice from a HollowSymmetricMatrix in compact form."""
    return _get_hsm_double_slice(hsm_compact, order, the_slice, slice(None))


def _get_hsm_col_slice(hsm_compact, order, the_slice):
    """Extract column slice from a HollowSymmetricMatrix in compact form."""
    return _get_hsm_double_slice(hsm_compact, order, slice(None), the_slice)


class _MemSize:
    """Prepare a "pretty-print" form of the HollowSymmetricMatrix memsize."""

    def __init__(self, size_in_bytes):
        if size_in_bytes < 0:
            raise ValueError("Arg 'size_in_bytes' cannot be negative")
        self.bytes = size_in_bytes

    def __repr__(self):
        return "{}(size_in_bytes={})".format(type(self).__name__, self.bytes)

    def __str__(self):
        return humanize.naturalsize(self.bytes)


# Dict mapping numpy funcs into our eventual custom implementations for them.
# To be edited via the decorator defined below.
_NUMPY_FUNCTIONS_HANDLED_BY_HSM = {}


# Define a decorator to allow us to define custom implementations of some numpy
# array functions. See <https://numpy.org/doc/stable/user/basics.dispatch.html>
def _implements_np_function_for_hsm(np_function):
    """Register __array_function__ implementation for HollowSymmetricMatrix."""

    def decorator(func):
        """Get decorator to add implementations of numpy array functions."""
        _NUMPY_FUNCTIONS_HANDLED_BY_HSM[np_function] = func
        return func

    return decorator


class HollowSymmetricMatrix(np.lib.mixins.NDArrayOperatorsMixin):
    """Hollow symmetric matrix with flexible memory storage options.

    A hollow symmetric matrix is a square matrix M of order n that satisfies:
        * M[i, j] = M[j, i] for all i, j in range(n)
        * M[i, i] = 0 for all i in range(n)

    There are therefore only (n**2 - n)/2 independent elements in such a
    matrix. A new instance of this class will store only the independent
    elements of a hollow symmetric matrix, while allowing access to such
    elements via regular matrix indexing.

    The order n of the underlying hollow symmetric matrix is inferred from
    the passed data array. A ValueError will be raised if the length of the
    data array is incompatible with the number of independent elements of a
    hollow symmetric matrix.

    There "optimize_mode" arg accepts three choices:
        * "speed"
            - The matrix will be stored in its dense (full) form
            - Copies will be returned as numpy 2D float64 arrays
        * "memory":
            - The matrix will be stored in compact form
            - Copies will return HollowSymmetricMatrix instances
        * "speed_mem_compromise":
            - The matrix will be stored in its dense (full) form
            - Copies will return HollowSymmetricMatrix instances
    When used with HDBSCAN, the following relation holds for memory usage:
                "memory" < "speed_mem_compromise" < "speed",
    and the inverse relation holds for execution times.

    """

    def __init__(self, data, dtype=None, optimize_mode="memory"):
        """Initialise data, optimize_mode and 'T' attrs."""
        self._set_data(data, dtype)

        # Decide how the matrix elements will be stored
        self._optimize_mode = optimize_mode.lower()
        opt_mode_choices = ["speed", "speed_mem_compromise", "memory"]
        if self._optimize_mode not in opt_mode_choices:
            msg = "'{}' not allowed for the 'optimize_mode' arg. "
            msg += "Please choose from: {}"
            msg = msg.format(optimize_mode, ", ".join(opt_mode_choices))
            raise ValueError(msg)
        if self._optimize_mode == "memory":
            self.convert_to_compact_storage()
        else:
            # self then becomes just a container for the 2D numpy array.
            # The difference between "speed" and "speed_mem_compromise"
            # modes lies in how self.copy behaves.
            self.convert_to_dense_storage()

        logger.debug("Init new %s object:\n%s", type(self).__name__, self)

    @classmethod
    def validated_data(cls, data):
        """Get HSM-compatible version of data. ValueError if not possible."""
        if np.ndim(data) == 1:
            n_indep = len(data)
            order = 0.5 * (1.0 + np.sqrt(1.0 + 8.0 * n_indep))
            nearest_int_order = int(np.rint(order))
            if abs(order - nearest_int_order) > 1e-8:
                nearest_n_indep = int(
                    0.5 * nearest_int_order * (nearest_int_order - 1)
                )
                msg = "len(data)={0} incompatible with {1}: {2} elements "
                msg += "needed for a {3}x{3} matrix (closest option)"
                msg = msg.format(
                    n_indep, cls.__name__, nearest_n_indep, nearest_int_order
                )
                raise ValueError(msg)
        elif np.ndim(data) == 2:
            if not np.equal(*np.shape(data)):
                raise ValueError("'data' is not a square matrix")
            if not np.allclose(data, np.transpose(data)):
                logger.warning(
                    "%s: Data not symmetric. Copying upper-diag into lower!",
                    cls.__name__,
                )
                # The diag should be zero, so we won't subtract it here.
                # If it isn't zero, then this will be taken care of next.
                triu = np.triu(data)
                data = triu + triu.T
            if any(data[_, _] != 0 for _ in range(np.shape(data)[0])):
                logger.warning(
                    "%s: Resetting diagonal elements to %s",
                    cls.__name__,
                    type(data[0, 0])(0),
                )
                # Note that np.fill_diagonal takes care of typecasting
                np.fill_diagonal(data, 0)
        else:
            raise ValueError(
                "%s: ndim(data) should be 1 or 2. Got %s."
                % (cls.__name__, np.ndim(data))
            )
        return data

    def _set_data(self, new, check_data_consistency=True, dtype=None):
        """Set/reset self._data."""
        if check_data_consistency:
            new = self.__class__.validated_data(new)
        if dtype is None:
            self._data = np.array(new)
        else:
            self._data = np.array(new, dtype=dtype)
        # The underlying data array should not be directly modified externally
        self._data.flags.writeable = False

    @property
    def optimize_mode(self):
        """Return the mode for which the matrix is optimised."""
        return self._optimize_mode

    @property
    def data_mem_usage(self):
        """Return mem usage of the data array contained within self."""
        return _MemSize(self._data.nbytes)

    @property
    def dtype(self):
        """Return the matrix data type."""
        return self._data.dtype

    @property
    def order(self):
        """Return the matrix order, as in "order N" for an NxN matrix."""
        if self.stored_as_compact:
            n_indep = len(self._data)
            rtn = int(np.rint(0.5 * (1.0 + np.sqrt(1.0 + 8.0 * n_indep))))
        else:
            rtn = self._data.shape[0]

        return rtn

    @property
    def n_indep_els(self):
        """Return the #independent (non-redundant off-diagonal) elements."""
        return (self.order ** 2 - self.order) // 2

    @property
    def shape(self):
        """Return the matrix shape."""
        return (self.order, self.order)

    @property
    def ndim(self):
        """Return the number of dimensions of the matrix."""
        return 2

    @_implements_np_function_for_hsm(np.transpose)
    def transpose(self):
        """Transpose of a HollowSymmetricMatrix object: A copy of the object.

        This also implements np.transpose for HollowSymmetricMatrix objects

        Returns:
            netatmoqc.hollow_symmetric_matrix.HollowSymmetricMatrix: Just a
                copy of self, as HollowSymmetricMatrix objects are, well,
                symmetric.

        """
        return self.copy()

    # Create a "T" property as an alias for self.transpose()
    T = property(transpose)

    @property
    def stored_as_compact(self):
        """Return True if matrix is stored in compact mode, False otherwise."""
        return self._data.ndim == 1

    @property
    def stored_as_dense(self):
        """Return True if matrix is stored in dense mode, False otherwise."""
        return not self.stored_as_compact

    def compact_form(self):
        """Return the compact-form representation of the matrix."""
        if self.stored_as_compact:
            rtn = self._data
        else:
            logger.debug("%s: Computing compact form", type(self).__name__)
            start = time.time()
            rtn = _to_compact(self._data, self.dtype)
            # Set the return value to read-only for consistency,
            # as this is how it would be if self.stored_as_compact
            # since self._data is read-only
            rtn.flags.writeable = False
            logger.debug(
                "Done computing compact form. Elapsed: %.2fs",
                time.time() - start,
            )

        return rtn

    def dense_form(self):
        """Return the dense-form representation of the matrix."""
        if self.stored_as_dense:
            rtn = self._data
        else:
            logger.debug(
                "%s: Computing %s dense form", type(self).__name__, self.shape
            )
            start = time.time()
            rtn = _to_dense(self._data, self.order, self.dtype)
            # Set the return value to read-only for consistency,
            # as this is how it would be if self.stored_as_dense
            # since self._data is read-only
            rtn.flags.writeable = False
            logger.debug(
                "Done computing dense form. Elapsed: %.2fs",
                time.time() - start,
            )

        return rtn

    def subspace(self, indices):
        """Subspace of self constructed as "irow in indices & icol=irow"."""
        logger.debug("%s: Subspace requested", type(self).__name__)

        # In case we get, e.g., panda Series objects
        indices = np.array(indices)

        # Fancy indexing as below is quicker than [indices, :][:, indices]
        if self.stored_as_compact:
            new_data = _get_hsm_subspace_from_compact(
                self.compact_form(), indices
            )
        else:
            new_data = self[indices[:, np.newaxis], indices]

        return self.__class__(
            new_data, dtype=self.dtype, optimize_mode=self.optimize_mode,
        )

    def convert_to_dense_storage(self):
        """Change matrix storage mode to "dense"."""
        if self.stored_as_dense:
            return
        logger.debug("%s: Switch to dense storage", type(self).__name__)
        self._set_data(self.dense_form(), check_data_consistency=False)

    def convert_to_compact_storage(self):
        """Change matrix storage mode to "compact"."""
        if self.stored_as_compact:
            return
        logger.debug("%s: Switch to compact storage", type(self).__name__)
        self._set_data(self.compact_form(), check_data_consistency=False)

    def index_dense_to_compact(self, i, j):
        """Return integer I such that self.data[I] == self[i, j]."""
        if i == j:
            raise HsmDiagNotStoredInCompactForm
        return _matrix_index_to_data_index(self.order, i, j)

    def index_compact_to_dense(self, data_i):
        """Return (i, j) such that self.data[data_i] == self[i, j]."""
        return _data_index_to_matrix_index(self.order, data_i)

    def __array__(self, dtype=None, copy=False, order=None):
        # See <https://numpy.org/devdocs/user/basics.dispatch.html>
        # TODO: Remove the ".astype(np.float64)" when possible.
        # HDBSCAN requires that the result of np.array(self) to have
        # dtype=numpy.float64. There's an open feature request at
        # <https://github.com/scikit-learn-contrib/hdbscan/issues/108>
        # (checked on 2020-09-02) for support to other types. Let's wait.
        logger.debug("%s: Calling '__array__'  method", type(self).__name__)
        return self.dense_form().astype(np.float64)

    def __getitem__(self, ij_tuple):
        if self.stored_as_dense:
            # Use numpy's own __getitem__
            rtn = self.dense_form()[ij_tuple]
        elif all(isinstance(index, int) for index in ij_tuple):
            # Just an element, neither slices nor fancy indexing
            try:
                i_data = self.index_dense_to_compact(*ij_tuple)
                rtn = self.compact_form()[i_data]
            except HsmDiagNotStoredInCompactForm:
                rtn = 0.0
        elif all(isinstance(_, slice) for _ in ij_tuple):
            logger.debug("%s: Get row and column slices", type(self).__name__)
            rtn = _get_hsm_double_slice(
                self.compact_form(), self.order, *ij_tuple
            )
        elif isinstance(ij_tuple[0], slice):
            logger.debug("%s: Get row slice", type(self).__name__)
            rtn = _get_hsm_row_slice(
                self.compact_form(), self.order, ij_tuple[0]
            )[:, ij_tuple[1]]
        elif isinstance(ij_tuple[1], slice):
            logger.debug("%s: Get column slice", type(self).__name__)
            rtn = _get_hsm_col_slice(
                self.compact_form(), self.order, ij_tuple[1]
            )[ij_tuple[0], :]
        else:
            logger.debug(
                "%s: Get fancy-indexed submatrix", type(self).__name__
            )
            bcasted_keys = _getitem_keys_to_bcast_arrays(ij_tuple)
            rtn = _get_submatrix_from_hsm_compact(
                self.compact_form(), self.order, bcasted_keys
            )

        return rtn

    def copy(self):
        """Return a copy of the matrix, formatted depending on optimize_mode.

        Returns:
            {
                numpy.ndarray: If optimize_mode == "speed". Dense form.
                netatmoqc.hollow_symmetric_matrix.HollowSymmetricMatrix: If
                    optimize_mode != "speed".
            }

        """
        logger.debug("%s: Calling 'copy'  method", type(self).__name__)
        if self.optimize_mode == "speed":
            # HDBSCAN calls this method multiple times, and then calls
            # np.asarray on the returned copy. If self is optimized for
            # speed, then let's return a numpy array already. This is going
            # to cost in memory, as the returned array is not simply a copy
            # of the already available dense form, but rather the dense
            # form casted into np.float64 (HDBSCAN requires this; see the
            # __array__ method).
            rtn = np.asarray(self)
        else:
            # In this case, __array_ufunc__ will be called if numpy ufuncs
            # are called with self as arg. Besides this:
            # * If optimize_mode == "speed_mem_compromise", then the dense
            # form is available (having thus an increased memory use w.r.t.
            # optimize_mode == "memory"), but it won't be cast into np.float64
            # unless np.asarray is called on self (thus reducing memory usage
            # w.r.t. optimize_mode == "speed").
            # * If optimize_mode == "memory", then the dense form is not kept
            # on memory and will be calculated only when necessary and kept
            # on memory only as long it is needed for. This saves memory but
            # increases computation.
            rtn = copy.copy(self)
        return rtn

    def __setitem__(self, ij_tuple, value):
        try:
            boolean_indexing = ij_tuple.dtype == np.bool
        except AttributeError:
            boolean_indexing = np.array(ij_tuple).dtype == np.bool

        if boolean_indexing:
            if isinstance(ij_tuple, type(self)):
                self._data.flags.writeable = True
                if self.stored_as_dense:
                    self._data[ij_tuple.dense_form()] = value
                else:
                    self._data[ij_tuple.compact_form()] = value
                self._data.flags.writeable = False
            else:
                initially_as_compact = self.stored_as_compact
                self.convert_to_dense_storage()
                self._data.flags.writeable = True
                self._data[ij_tuple] = value
                # Using self._set_data to check for invalid values
                self._set_data(self._data)
                if initially_as_compact:
                    self.convert_to_compact_storage()
        elif all(isinstance(_, (int, np.integer)) for _ in ij_tuple):
            # Just an element, neither slices nor fancy indexing
            self._data.flags.writeable = True
            try:
                i_data = self.index_dense_to_compact(*ij_tuple)
                if self.stored_as_dense:
                    # Make sure the matrix remains symmetric
                    self._data[ij_tuple] = value
                    self._data[ij_tuple[::-1]] = value
                else:
                    self._data[i_data] = value
            except HsmDiagNotStoredInCompactForm:
                # Make sure the matrix remains hollow
                logger.warning(
                    "Cannot set %s diagnonal elements. Doing nothing.",
                    type(self).__name__,
                )
            self._data.flags.writeable = False
        else:
            msg = "Setting {} items is only supported via bool or int indices"
            raise NotImplementedError(msg.format(type(self).__name__))

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        ufunc_method = getattr(ufunc, method)
        if all(isinstance(_, self.__class__) for _ in inputs):
            logger.debug(
                "%s: __array_ufunc__ handling same-class '%s.%s'",
                type(self).__name__,
                ufunc.__name__,
                method,
            )

            # Test if the ufunc still returns a hollow symmetric matrix
            try:
                test_hsm = np.array([0, 1, 1, 0]).reshape(2, 2)
                test_inputs = tuple(test_hsm.astype(_.dtype) for _ in inputs)
                test_result = ufunc_method(*test_inputs, **kwargs)
                result_is_hsm = (
                    # Check 1: Result is a matrix
                    test_result.ndim == 2
                    # Check 2: Result matrix is square
                    and np.equal(*test_result.shape)
                    # Check 3: Result matrix is symmetric
                    and np.allclose(test_result, test_result.T)
                    # Check 2: Result matrix diagonal is zero
                    and np.allclose(
                        ufunc_method(*test_inputs, **kwargs).diagonal(),
                        test_hsm.diagonal().astype(test_result.dtype),
                    )
                )
            except (ValueError, AttributeError):
                result_is_hsm = False

            if result_is_hsm and any(_.stored_as_compact for _ in inputs):
                new_inputs = tuple(_.compact_form() for _ in inputs)
            else:
                new_inputs = tuple(_.dense_form() for _ in inputs)
            result = ufunc_method(*new_inputs, **kwargs)

            if result_is_hsm:
                # Return another instance of type(self)
                if result.ndim != 1:
                    result = _to_compact(result, dtype=result.dtype)
                return self.__class__(
                    result,
                    dtype=result.dtype,
                    optimize_mode=self.optimize_mode,
                )
            else:
                # Return whatever the ufunc call returns
                logger.debug(
                    "ufunc '%s.%s(%s)' returning '%s', not '%s'",
                    ufunc.__name__,
                    method,
                    ", ".join(type(_).__name__ for _ in inputs),
                    type(result).__name__,
                    type(self).__name__,
                )
                return result
        else:
            # Act on the dense form, which is just a regular 2D numpy array
            logger.debug(
                "%s: __array_ufunc__ handling mixed-class '%s.%s'",
                type(self).__name__,
                ufunc.__name__,
                method,
            )
            new_inputs = [
                np.asarray(inp) if isinstance(inp, self.__class__) else inp
                for inp in inputs[:2]
            ]
            return ufunc_method(*new_inputs, **kwargs)

    def __array_function__(self, func, types, args, kwargs):
        """Allow class to define how numpy functions operate on its objects."""
        # Adapted from Numpy's own suggestions found on:
        # <https://numpy.org/neps/nep-0018-array-function-protocol.html>
        # <https://numpy.org/doc/stable/user/basics.dispatch.html>

        # Allow subclasses that don't override __array_function__
        # to handle HollowSymmetricMatrix objects
        if not all(issubclass(t, (self.__class__, np.ndarray)) for t in types):
            return NotImplemented

        try:
            # Use our custom implementation, if available
            array_func = _NUMPY_FUNCTIONS_HANDLED_BY_HSM[func]
            logger.debug(
                "%s (%s): Using custom implementation '%s' for numpy's '%s'",
                type(self).__name__,
                "__array_function__",
                array_func.__name__,
                func.__name__,
            )
        except KeyError:
            # Use NumPy's private implementation without __array_function__
            # dispatching if we don't have a custom implementation
            array_func = func._implementation
            logger.debug(
                "%s (%s): Using numpy's own implementation of '%s'",
                type(self).__name__,
                "__array_function__",
                func.__name__,
            )
        return array_func(*args, **kwargs)

    def __repr__(self):
        txt = "%s(\n"
        txt += "    shape=%s\n"
        txt += "    dtype=%s\n"
        txt += "    optimize_mode=%s\n"
        if self.stored_as_dense:
            txt += "    storage='dense' (%s elements stored in memory)\n"
            txt += "    memsize='%s'\n"
            txt += "    data:\n%s\n"
            data = self.dense_form()
        else:
            txt += "    storage='compact' (%s elements stored in memory)\n"
            txt += "    memsize=%s\n"
            txt += "    data=%s\n"
            data = self.compact_form()
        txt += ")"
        return txt % (
            type(self).__name__,
            self.shape,
            self.dtype,
            self.optimize_mode,
            data.size,
            self.data_mem_usage,
            data,
        )

    def __str__(self):
        order = self.order
        return self.dense_form().__str__() if order < 5 else self.__repr__()



#!/usr/bin/env python3
"""Routines, classes and definitions employed to load NetAtmo data."""
import logging
import time
from datetime import datetime
from pathlib import Path

import humanize
import numpy as np
import pandas as pd
from pytz import timezone

from .dtgs import Dtg, epoch2datetime

logger = logging.getLogger(__name__)


class DataNotFoundError(Exception):
    """Exeption to be raised when data has not been found."""


class DataFrameWithDtgMetadata(pd.DataFrame):
    """Pandas DataFrame with added "metadata_dict" attribute."""

    # See <https://pandas.pydata.org/pandas-docs/stable/development/
    #      extending.html#define-original-properties>
    # Define _internal_names and _internal_names_set for temporary properties
    # which WILL NOT be passed to manipulation results.
    # Define _metadata for normal properties which will be passed to
    # manipulation results.
    # See also:
    # <https://stackoverflow.com/a/23200907>
    # <https://github.com/pandas-dev/pandas/issues/6927#issuecomment-41086742>

    # temporary properties
    _internal_names = pd.DataFrame._internal_names + ["temp_property"]
    _internal_names_set = set(_internal_names)

    # normal properties
    _metadata = pd.DataFrame._metadata + ["metadata_dict"]

    def __init__(self, *args, **kwargs):
        """Initialise pandas DataFrame and then add "metadata_dict" attr."""
        self.metadata_dict = {}
        super().__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return DataFrameWithDtgMetadata

    def __finalize__(self, other, method=None, **kwargs):
        for name in self._metadata:
            if name == "metadata_dict":
                # Sometimes "other" is actually an iterable. We'll make it
                # so the final dict merge works with the iterable "others"
                try:
                    _ = other.metadata_dict
                    # Has metadata_dict attr. Let's use it as "others"
                    others = [other]
                except AttributeError:
                    try:
                        # "other" has "objs" attr. Let "objs" become "others".
                        others = other.objs
                    except AttributeError:
                        # No "objs" attr. Assume "other" is itself iterable and
                        # allow TypeError to be raised later otherwise (because
                        # then we don't know what "other" is)
                        others = other

                # Compare metadata dicts and keep only common keys/vals
                md_dicts_to_combine = [getattr(_, name, {}) for _ in others]
                self_md_dict = getattr(self, name, {})
                if self_md_dict:
                    md_dicts_to_combine.append(self_md_dict)

                common_keys = set()
                for dic in md_dicts_to_combine:
                    if dic == {}:
                        # If any dict is empty, then there will be no
                        # common keys at all.
                        common_keys = set()
                        break
                    common_keys = common_keys.union(set(dic))

                combined_md_dict = {}
                for key in common_keys:
                    val = np.unique([dic[key] for dic in md_dicts_to_combine])
                    if len(val) == 1:
                        combined_md_dict[key] = val[0]

                object.__setattr__(self, name, combined_md_dict)
            else:
                object.__setattr__(self, name, getattr(other, name, None))

        return self


@np.vectorize
def mslp2sp_coeff(alt):
    """Coeff to convert mean sea-level pressure to pressure at altitude alt.

    Args:
        alt: Altitude in meters.

    Returns:
        float: Which multiplies by the mslp to give the surface pressure at alt

    """
    return 1.0 / (
        int(round(100000 / ((288 - 0.0065 * alt) / 288) ** 5.255)) / 100000
    )


def remove_str_prefix(string, prefix):
    """Remove prefix from string."""
    if string.startswith(prefix):
        rtn = string[len(prefix) :]
    else:
        rtn = string
    return rtn


@np.vectorize
def shorten_stat_id(stat_id):
    """Shorten station ID to length 8."""
    if not isinstance(stat_id, str):
        return stat_id
    # Try first to produce IDs consistent with what Roel does in Netatmo.jl,
    # to be comparable, as there's already people using Netatmo.jl.
    new_id = stat_id[9:17]
    if len(new_id) != 8:
        # Use the strategy originally adapted in netatmoqc if the new ID
        # is shorter than expected
        new_id = remove_str_prefix(stat_id, prefix="enc:16:")
    return new_id[:8]


def read_netatmo_csv(
    path,
    dropna=True,
    fillna=None,
    recover_pressure_from_mslp=False,
    drop_mslp=None,
):
    """Read CSV files with NetAtmo observation data.

    Args:
        path (pathlib.Path): Full path to the csv file.
        dropna (bool): Whether or not to exclude NaNs (Default value = True).
        fillna (dict): Dictionary having data parameters as keys and the fill
            values (which will replace NaNs) as dict values. The filling of
            NaNs is performed prior to dropping NaNs. (Default value = None).
        recover_pressure_from_mslp (bool): The netatmo csv files are assumed to
            contain mslp data instead of surface pressure. This flag controls
            whether surface pressure should also be included in the read data.
            If True. then a new column "pressure" will be added with surface
            pressure calculated from altitude and mslp.
            (Default value = False)
        drop_mslp (bool): Whether ot not to drop the original mslp after
            having created a new "pressure" column with surface pressure.
            If None is used, then drop_mslp takes the value of the
            recover_pressure_from_mslp parameter.
            (Default value = None)

    Returns:
        pandas.Dataframe: Data read from the input netatmo csv file.

    Raises:
        ValueError: If drop_mslp evaluates to True while
            recover_pressure_from_mslp evaluates to False.

    """
    if drop_mslp is None:
        drop_mslp = recover_pressure_from_mslp
    if drop_mslp and not recover_pressure_from_mslp:
        raise ValueError(
            "Cannot have drop_mslp=True and recover_pressure_from_mslp=False"
        )

    # Set fillna to default value if None passed.
    # Do this here instead of having a dict default because dict is mutable,
    # and default args are created at function definition time.
    if fillna is None:
        fillna = dict(sum_rain_1=0.0)

    data = pd.read_csv(path)
    # Replacing/removing invalid data (NaNs)
    if fillna:
        # fillna may be a key:val map (dict), in which case NaNs in col "key"
        # will be filled with value "val". Fill NaNs before dropping!
        data = data.fillna(fillna)
    if dropna:
        data = data.dropna()

    # Drop the 'enc:16:' stat id prefix and shorten them to 8 chars
    data["id"] = shorten_stat_id(data["id"])

    # The netatmo data I got from Norway has mean sea-level pressure
    # instead of pressure, but the label is "pressure" there. Fix this.
    data = data.rename(columns={"pressure": "mslp"})
    pressure_cols = ["mslp"]
    if recover_pressure_from_mslp:
        data["pressure"] = mslp2sp_coeff(data["alt"]) * data["mslp"]
        pressure_cols += ["pressure"]
    # Make sure columns 'id', 'time_utc', 'lat' and 'lon' appear
    # first AND in this order
    data = data[
        ["id", "time_utc", "lat", "lon", "alt"]
        + pressure_cols
        + ["temperature", "humidity", "sum_rain_1"]
    ]
    if drop_mslp:
        data = data.drop(["mslp"], axis=1)

    data["time_utc"] = epoch2datetime(data["time_utc"])

    return data.sort_values(by="time_utc").reset_index(drop=True)


def gen_netatmo_fpaths_for_dates(dates, rootdir):
    """Generate netatmo file paths for dates.

    NetAtmo CSV files are assumed to have full paths following the pattern
    "rootdir/%Y/%m/%d/%Y%m%dT%H%M%SZ.csv"

    Args:
        dates (list): List of datetime.datetime objects. Only year, month, and
            day are used.
        rootdir (pathlib.Path): Path to the top-level directory where the
            netatmo input csv files are stored.

    Yields:
        libpath.Path: Paths to files be read in order to get data for date=dt

    """
    try:
        date_iter = iter(dates)
    except TypeError:
        date_iter = [dates]
    rootdir = Path(rootdir).resolve()
    for date in sorted(date_iter):
        year = date.strftime("%Y")
        month = date.strftime("%m")
        day = date.strftime("%d")
        for fpath in sorted((rootdir / year / month / day).glob("*.csv")):
            yield fpath


def read_netatmo_data_for_dates(dates, rootdir, **kwargs):
    """Read netatmo data for passed dates.

    Args:
        dates (list): List of datetime.datetime objects. Only year, month, and
            day are used.
        rootdir (pathlib.Path): Path to the top-level directory where the
            netatmo input csv files are stored.
        **kwargs: Passed to read_netatmo_csv

    Returns:
        pandas.Dataframe: Data read from csv files for the dates passed
            as input.

    Raises:
        DataNotFoundError: If data for the selected dates cannot be found.

    """
    data_from_all_files = {}
    for fpath in gen_netatmo_fpaths_for_dates(dates, rootdir):
        data_from_all_files[fpath.stem] = read_netatmo_csv(fpath, **kwargs)

    if len(data_from_all_files) == 0:
        raise DataNotFoundError(
            "Could not find data for date(s)={} under dir '{}'".format(
                dates, rootdir
            )
        )

    return pd.concat(data_from_all_files, ignore_index=True)


def _datetime_of_netatmo_file(fpath, fmt="%Y%m%dT%H%M%SZ.csv"):
    """Get the datetime from csv file path."""
    dt = datetime.strptime(fpath.name, fmt)
    return dt.replace(tzinfo=timezone("utc"))


# Reading data for a given DTG instead of date
def gen_netatmo_fpaths_for_time_window(start, end, rootdir):
    """Generate NetAtmo csv file paths for datetimes within [start, end].

    Returns a generator of paths leading to NetAtmo files whose names
    correspond to timestamps lying within the closed interval [start, end].

    Args:
        start (datetime.datetime): Start date.
        end (datetime.datetime): End date.
        rootdir (pathlib.Path): Path to the top-level directory
            where the netatmo input csv files are stored.

    Yields:
        pathlib.Path: NetAtmo csv file paths.

    """
    start, end = sorted((start, end))
    dates = pd.date_range(start.date(), end.date(), freq="1d")
    for fpath in gen_netatmo_fpaths_for_dates(dates, rootdir):
        date_time = _datetime_of_netatmo_file(fpath)
        if start <= date_time <= end:
            yield fpath


def gen_netatmo_fpaths_for_dtg(dtg, rootdir):
    """Paths to NetAtmo files that are likely to contain data for DTG=dtg.

    Args:
        dtg (netatmoqc.dtgs.Dtg): Data assimilation DTG.
        rootdir (pathlib.Path): Path to the top-level directory
            where the netatmo input csv files are stored.

    Returns:
        generator: Generator of the NetAtmo csv file paths

    """
    # Add a few minutes to the assimilation window, as
    # netatmo data is gathered every ~10 minutes
    start = dtg.cycle_start - pd.Timedelta("15 minutes")
    end = dtg.cycle_end + pd.Timedelta("15 minutes")
    return gen_netatmo_fpaths_for_time_window(start, end, rootdir=rootdir)


def rm_moving_stations(df):
    """Remove stations that change (lon, lat, alt) from the data.

    Args:
        df (pandas.Dataframe): Input data.

    Returns:
        {
            'pandas.Dataframe': Copy of input df, but with
                moving stations removed,
            'pandas.Dataframe': The moving stations.
        }

    """
    df_grouped_by_id = df.groupby(["id"], as_index=False, sort=False)
    df_temp = df_grouped_by_id[["lat", "lon", "alt"]].agg(["max", "min"])
    inconsistent_stations = df_temp[
        (abs(df_temp["lat"]["max"] - df_temp["lat"]["min"]) >= 1e-4)
        | (abs(df_temp["lon"]["max"] - df_temp["lon"]["min"]) >= 1e-4)
        | (abs(df_temp["alt"]["max"] - df_temp["alt"]["min"]) >= 1e-2)
    ].index.values
    df_ok = df[~df["id"].isin(inconsistent_stations)].reset_index(drop=True)
    df_inconsistent = df[df["id"].isin(inconsistent_stations)].reset_index(
        drop=True
    )
    return df_ok, df_inconsistent


def rm_overlapping_stations(df):
    """Remove from df the stations that have the same (lon, lat).

    Args:
        df (pandas.Dataframe): Input data.

    Returns:
        'pandas.Dataframe': Copy of input df, but with the
            overlapping stations removed.

    """
    overlapping_stations = (
        df[["id", "lat", "lon"]]
        .round(6)
        .groupby(["lat", "lon"], as_index=False, sort=False,)
        .filter(lambda grp: len(grp["id"].unique()) != 1)["id"]
        .unique()
    )
    return df[~df["id"].isin(overlapping_stations)].reset_index(drop=True)


def remove_duplicates_within_cycle(df, dtg):
    """Remove duplicate station reports within DTG.

    Remove duplicates for each station by keeping the one closest to DTG

    If a station reports data multiple times within the time window,
    then keep the one closest to the passed DTG and discard the rest.

    Args:
        df (pandas.Dataframe): Input data.
        dtg (netatmoqc.dtgs.Dtg): The picked assimilation DTG.

    Returns:
        'pandas.Dataframe': Copy of input df, where only one station
            report is kept for each DTG.

    """
    # It seems that using groupby followed by "apply" with a custom
    # function is about 20x slower.
    # Sort using df.loc is better than adding a new col to sort by it. See
    # <stackoverflow.com/questions/39525928/pandas-sort-lambda-function>
    df = df.loc[(abs(df["time_utc"] - dtg)).sort_values().index]
    # We don't need sorting with groupby. HDBSCAN clustering, however,
    # seems to be sensitive to the order in which the rows appear (but
    # not very much so). This seems to be knwon, see, e.g.,
    # <https://github.com/scikit-learn-contrib/hdbscan/issues/265>
    station_groups = df.groupby(["id"], as_index=False, sort=False)
    n_stations_with_duplicates = (station_groups.size()["size"] > 1).sum()
    if n_stations_with_duplicates > 0:
        orig_nobs = len(df.index)
        df = station_groups.first()
        new_nobs = len(df.index)
        msg = "Multiple reports from %d stations in DTG=%s. "
        msg += "Keeping only the one closest to the DTG: "
        msg += "%s obs now became %s"
        logger.debug(
            msg, n_stations_with_duplicates, dtg, orig_nobs, new_nobs,
        )
    return df


def read_netatmo_data_for_dtg(dtg, rootdir, **kwargs):
    """Read NetAtmo data corresponding to the selected DTG.

    Args:
        dtg (netatmoqc.dtgs.Dtg): The picked assimilation DTG.
        rootdir (pathlib.Path): Path to the top-level directory where the
            netatmo input csv files are stored.
        **kwargs: Passed to read_netatmo_csv

    Returns:
        pandas.Dataframe: Data read from csv files for the input DTG.

    Raises:
        DataNotFoundError: If data for the selected DTGs cannot be found.

    """
    t_start = time.time()
    logger.debug("Reading data for DTG=%s...", dtg)

    # Make sure the input DTG behaves as expected
    dtg = Dtg(dtg)

    data_from_all_files = {}
    for fpath in gen_netatmo_fpaths_for_dtg(dtg, rootdir=rootdir):
        data_from_all_files[fpath.stem] = read_netatmo_csv(fpath, **kwargs)
    if len(data_from_all_files) == 0:
        raise DataNotFoundError(
            "Could not find data for DTG={} under dir '{}'".format(
                dtg, rootdir
            )
        )

    df = pd.concat(data_from_all_files, ignore_index=True)
    # Remove DTGs that do not belong to the assimilation time window
    if dtg.assimilation_window.closed_left:
        mask_left = df["time_utc"] >= dtg.assimilation_window.left
    else:
        mask_left = df["time_utc"] > dtg.assimilation_window.left
    if dtg.assimilation_window.closed_right:
        mask_right = df["time_utc"] <= dtg.assimilation_window.right
    else:
        mask_right = df["time_utc"] < dtg.assimilation_window.right
    df = df[mask_left & mask_right]

    # Add DTG metadata, for a future reference of how the data was collated
    df = DataFrameWithDtgMetadata(df)
    df.metadata_dict["dtg"] = dtg

    logger.debug(
        "Dataframe has obs from %d stations, with a memsize=%s",
        len(df.index),
        humanize.naturalsize(df.memory_usage(deep=True).sum()),
    )

    logger.debug(
        "Done reading data for DTG=%s. Elapsed: %.1fs",
        dtg,
        time.time() - t_start,
    )

    return df


def remove_irregular_stations(
    df, moving=True, duplicates=True, overlapping=True
):
    """Remove obs from irregular stations (moving, duplicates or overlaps).

    This is a wrapper for the following routines:
        * rm_moving_stations
        * remove_duplicates_within_cycle
        * rm_overlapping_stations

    Args:
        df (pandas.Dataframe): Input data.
        moving (bool): Whether or not to remove moving stations.
            (Default value = True)
        duplicates (bool): Whether or not to remove duplicate sttaion reports.
            Default value = True)
        overlapping (bool): Whether or not to remove overlapping stations.
            Default value = True)

    Returns:
        {
            'pandas.Dataframe': Copy of input df, but with the
                irregular. stations removed,
            'pandas.Dataframe': Dataframe containing removed movind stations.
        }

    """
    dtg = df.metadata_dict["dtg"]
    if moving:
        # Remove stations that change (lat, lon, alt) within cycle
        df, df_moving_stations = rm_moving_stations(df)
        if len(df_moving_stations.index) > 0:
            logger.debug(
                'Removed %d obs from %d "moving" stations in DTG=%s',
                len(df_moving_stations.index),
                len(df_moving_stations["id"].unique()),
                dtg,
            )
    else:
        df_moving_stations = pd.DataFrame.empty
        logger.warning("Not checking for moving stations")

    if duplicates:
        # Remove duplicate data for stations within cycle by keeping
        # the data reported for the timestamp closest to the passed DTG
        df = remove_duplicates_within_cycle(df, dtg)
    else:
        logger.warning("Not removing duplicate stations")

    if overlapping:
        df = rm_overlapping_stations(df)

    df.metadata_dict["dtg"] = dtg

    return df, df_moving_stations



#!/usr/bin/env python3
"""Wrappers for argparse functionality."""
import argparse
import os
import subprocess
from datetime import datetime
from pathlib import Path

from . import __version__
from .commands_functions import (
    cluster_obs_single_dtg,
    csv2obsoul,
    select_stations,
    show,
)


class StoreDictKeyPair(argparse.Action):
    """Enable args="key1=val1, ..., keyN=valN" in command line args."""

    # Source: <https://stackoverflow.com/questions/29986185/
    #          python-argparse-dict-arg/42355279>

    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        """Initialise nargs."""
        self._nargs = nargs
        super().__init__(option_strings, dest, nargs=nargs, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        """Get a key/value dict from passed string."""
        my_dict = {}
        for k_v in values:
            key, val = k_v.split("=")
            my_dict[key] = val
        setattr(namespace, self.dest, my_dict)


def get_parsed_args(program_name):
    """Get parsed command line arguments.

    Args:
        program_name (str): The name of the program.

    Returns:
        argparse.Namespace: Parsed command line arguments.

    """
    # Define main parser and general options
    parser = argparse.ArgumentParser(
        prog=program_name,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    try:
        fpath = Path(os.getenv("NETATMOQC_CONFIG_PATH", "config.toml"))
        default_conf_path = fpath.resolve(strict=True)
    except FileNotFoundError:
        default_conf_path = (
            Path(os.getenv("HOME")) / ".netatmoqc" / "config.toml"
        )
    parser.add_argument(
        "--version", "-v", action="version", version="%(prog)s v" + __version__
    )
    parser.add_argument(
        "-config_file",
        metavar="CONFIG_FILE_PATH",
        default=default_conf_path,
        type=Path,
        help=(
            "Path to the config file. The default is whichever of the "
            + "following is first encountered: "
            + "(i) The value of the 'NETATMOQC_CONFIG_PATH' envvar or "
            + "(ii) './config.toml'. If both (i) and (ii) are missing, "
            + "then the default will become "
            + "'"
            + str(Path("$HOME/.netatmoqc/config.toml"))
            + "'"
        ),
    )
    parser.add_argument(
        "-loglevel",
        default="info",
        choices=["critical", "error", "warning", "info", "debug", "notset"],
        help="What type of info should be printed to the log",
    )
    parser.add_argument(
        "--mpi",
        action="store_true",
        help=(
            "Enable MPI parallelisation in some parts of the code. "
            + "Requires that the code be installed with support to MPI."
        ),
    )

    # Configure the main parser to handle the commands
    # From python v3.7, add_subparsers accepts a 'required' arg. We
    # could use required=True in the definition below, but, since we
    # want to be able to use python >= 3.6.10, we won't. The requirement
    # of having a command passed will be enforced in the main.py file.
    subparsers = parser.add_subparsers(
        title="commands",
        dest="command",
        description=(
            "Valid commands for {0} (note that commands also accept their "
            + "own arguments, in particular [-h]):"
        ).format(program_name),
        help="command description",
    )

    ##############################################
    # Configure parser for the "cluster" command #
    ##############################################
    parser_cluster = subparsers.add_parser(
        "cluster", help="Run cluster on NetAtmo obs for a single DTG"
    )
    parser_cluster.add_argument(
        "-dtg",
        type=lambda s: datetime.strptime(s, "%Y%m%d%H"),
        default=None,
        help="""
        If not passed, then the first DTG in the config file will be used.
        """,
    )
    parser_cluster.add_argument(
        "--show",
        action="store_true",
        help="Open a browser window to show results on a map",
    )
    parser_cluster.add_argument(
        "--savefig",
        action="store_true",
        help="Save an html file showing results on a map",
    )
    parser_cluster.set_defaults(func=cluster_obs_single_dtg)

    #############################################
    # Configure parser for the "select" command #
    #############################################
    parser_select = subparsers.add_parser(
        "select",
        help=(
            "Go over a list of DTGs and use clustering to select NetAtmo "
            + "stations that are reliable throughout the survey"
        ),
    )
    parser_select.add_argument(
        "--save-obsoul",
        action="store_true",
        help="Re-read the input NetAtmo files and save parsed data in OBSOUL "
        + "format, keeping only data related to the the accepted stations",
    )
    parser_select.set_defaults(func=select_stations)

    #################################################
    # Configure parser for the "csv2obsoul" command #
    #################################################
    parser_csv2obsoul = subparsers.add_parser(
        "csv2obsoul",
        help="Convert data from NetAtmo CSV files into OBSOUL format",
    )
    parser_csv2obsoul.add_argument(
        "--rm-duplicate-stations",
        action="store_true",
        help="Remove duplicates of station reports within same DTG. It keeps "
        + "the one whose obs time is closest to the asismilation time",
    )
    parser_csv2obsoul.add_argument(
        "--rm-moving-stations",
        action="store_true",
        help="Remove stations that change any of (lat, lon, alt) within the "
        + "selected asismilation windows",
    )
    parser_csv2obsoul.add_argument(
        "--dropna",
        action="store_true",
        help="Drop NA (NaN, not a number) values",
    )
    parser_csv2obsoul.add_argument(
        "--fillna",
        # 0.1699999976e39 is obsoul's sentinel value
        default=0.1699999976e39,
        action=StoreDictKeyPair,
        nargs="+",
        metavar="data_column=fillna_value_for_col",
    )
    parser_csv2obsoul.add_argument(
        "--selected-stations-fpath",
        metavar="STATIONS_FILE_PATH",
        default=None,
        type=Path,
        help=(
            "Path to the optional selected stations file. If specified, "
            + "then only these stations are kept."
        ),
    )

    parser_csv2obsoul.set_defaults(func=csv2obsoul)

    ###########################################
    # Configure parser for the "apps" command #
    ###########################################
    parser_apps = subparsers.add_parser(
        "apps",
        help=(
            "Start a server for the provided Dash apps. "
            + "N.B.: The apps are meant to be used for tests only."
        ),
    )
    parser_apps.set_defaults(
        func=lambda args: subprocess.run(
            [
                Path(__file__).parent.resolve() / "bin" / "start_apps",
                args.config_file.resolve(),
            ],
            check=True,
        )
    )

    ###########################################
    # Configure parser for the "show" command #
    ###########################################
    parser_show = subparsers.add_parser(
        "show",
        help="Display results from netatmoqc output files, as well as configs",
    )
    parser_show.add_argument(
        "file_list",
        nargs="*",
        type=Path,
        default=list(Path(".").glob("*.csv")),
        help="List of files the results should be read from",
    )
    parser_show.add_argument(
        "--config",
        action="store_true",
        help="Dump the used configs, including defaults, in TOML format",
    )
    parser_show.add_argument(
        "--domain",
        action="store_true",
        help="Open a browser window and show the configured domain",
    )
    parser_show.set_defaults(func=show)

    args = parser.parse_args()

    return args



#!/usr/bin/env python3
"""Implement Dtg object, Dtgs container, and related routines."""
import logging
from datetime import datetime, timedelta
from functools import partial

import numpy as np
import pandas as pd
from pandas.tseries.frequencies import to_offset
from pytz import timezone

logger = logging.getLogger(__name__)


@np.vectorize
def epoch2datetime(epoch):
    """Convert unix epoch into datetime object."""
    return datetime.fromtimestamp(epoch, timezone("utc"))


@np.vectorize
def datetime2epoch(dt):
    """Convert datetime object into unix epoch."""
    reference = datetime(1970, 1, 1, tzinfo=timezone("utc"))
    epoch = int((dt - reference).total_seconds())
    return epoch


class Dtg(datetime):
    """Implement Dtg object: NWP data assimilation time window."""

    def __init__(self, *args, **kwargs):
        """Set cycle_length attribute.

        Args:
            *args: Positional args passed to datetime.
            *kwargs: Keyword args passed to datetime.

        Attributes:
            cycle_length (str): The length of the assimilation window.

        """
        # Set a default of 3H for the length of a cycle.
        # Without the context provided by a cycle length, a
        # DTG has no meaning.
        self.set_cycle_length(kwargs.get("cycle_length", "3H"))

    def __new__(cls, *args, **kwargs):
        """Create new instance keeping only Dtg-compatible datetime attrs."""
        # cycle_length checks will only be performed in __init__
        kwargs.pop("cycle_length", None)

        # timezone will be set to UTC if not passed
        try:
            tzinfo = args[7]
            args = args[:7]
        except IndexError:
            tzinfo = kwargs.pop("tzinfo", timezone("utc"))
        if isinstance(tzinfo, str):
            tzinfo = timezone(tzinfo)
        if "tzinfo" not in kwargs:
            kwargs["tzinfo"] = tzinfo

        # Make it possible to create instance using either string or an
        # instance of datetime as the 1st positional argument. Keep support
        # for the datetime instantiation behaviour (with integers for year,
        # month, day, etc).
        if len(args) > 0 and not isinstance(args[0], int):
            if not any(isinstance(args[0], c) for c in [str, datetime]):
                raise TypeError(
                    "Cannot convert input '{}' ".format(args[0])
                    + "of type '{}' ".format(type(args[0]).__name__)
                    + "to {}".format(cls.__name__)
                )

            if len(args) > 1:
                raise ValueError(
                    "Only one positional argument is allowed "
                    + "if the first argument is not an integer"
                )

            try:
                # Add support to the YYYYMMDDHH standard DTG format
                # Assumed to be in UTC unless tzinfo is specified
                dt = datetime.strptime(args[0], "%Y%m%d%H")
                dt = dt.replace(tzinfo=kwargs["tzinfo"])
            except (ValueError, TypeError):
                # pd.Timestamp is flexible with string and datetime-like args
                # It is more convenient using pd.Timestamp than trying to
                # guess the type and format of args[0] manually
                ts_kwargs = kwargs.copy()
                try:
                    if args[0].tzinfo is not None:
                        ts_kwargs.pop("tzinfo")
                except AttributeError:
                    pass
                dt = pd.Timestamp(*args, **ts_kwargs).to_pydatetime()
                dt = dt.astimezone(kwargs["tzinfo"])

            # Clear args and pass only kwargs to datetime's __new__
            args = ()
            for att in [
                "year",
                "month",
                "day",
                "hour",
                "minute",
                "second",
                "microsecond",
                "tzinfo",
            ]:
                kwargs[att] = getattr(dt, att)

        new_instance = super().__new__(cls, *args, **kwargs)
        # Make sure DTGs don't have subdivisions smaller than one hour
        # This can easily happen if the instance is created based on string
        # or datetime-like arguments, or when it is created as a result of an
        # addition/subtraction operation
        eqv_valid_instance = super().__new__(
            cls,
            new_instance.year,
            new_instance.month,
            new_instance.day,
            new_instance.hour,
            tzinfo=new_instance.tzinfo,
        )
        if new_instance.as_datetime() != eqv_valid_instance.as_datetime():
            msg = "Dtg {} not allowed: ".format(new_instance.as_datetime())
            msg += "Minimum allowed Dtg subdivision is an hour"
            raise ValueError(msg)

        return new_instance

    def __reduce_ex__(self, protocol):
        # We define this method so that pickle knows how to recover a pickled
        # object of this class. If we don't do this, then pickle tries to pass
        # a bytes object to the classe's __new__ method upon inpickling. This
        # also indirectly affects copy and deepcopy. See a similar issue at
        # <https://stackoverflow.com/questions/59571625/
        #  unable-to-pickle-datetime-subclass>
        # Using functools.partial to be able to pass kwargs
        callable_for_new = partial(type(self), cycle_length=self.cycle_length)
        args = (self.isoformat(),)
        return callable_for_new, args

    # Handy type conversions
    def as_pandas_timestamp(self):
        """Return a copy as a panda's Timestamp object."""
        return pd.Timestamp(self)

    def as_datetime(self):
        """Return a copy as a datetime.datetime object."""
        return self.as_pandas_timestamp().to_pydatetime()

    # Useful little methods/properties
    @property
    def cycle_length(self):
        """Return the length of the data assimilation cycle."""
        return self._cycle_length

    @property
    def cycle_start(self):
        """Return the earliest observation time encompassed by the DTG."""
        return self.as_pandas_timestamp() - 0.5 * self.cycle_length.delta

    @property
    def cycle_end(self):
        """Return the first observation time not encompassed by the DTG."""
        return self.as_pandas_timestamp() + 0.5 * self.cycle_length.delta

    @property
    def assimilation_window(self):
        """Return the DTG's assimilation window: [cycle_start, cycle_end)."""
        return pd.Interval(self.cycle_start, self.cycle_end, closed="left")

    def get_next(self):
        """Return the next DTG."""
        return self + self.cycle_length

    def get_previous(self):
        """Return the previous DTG."""
        return self - self.cycle_length

    # cycle_length-related methods
    def compatible_with_cycle_length(self, cycle_length="self"):
        """Check if DTG is compatible with cycle length.

        Args:
            cycle_length (str): A cycle length. Default value = "self".

        Returns:
            bool: True if DTG is compatible with cycle length else False.

        """
        if cycle_length is None:
            return True

        if cycle_length == "self":
            cycle_length = self.cycle_length
        else:
            cycle_length = to_offset(cycle_length)
        # It is convenient to check this using pandas Timestamp objects
        # At the time of writing this, inheriting from pd.Timestamp lead
        # to strange behaviours.
        dtg = pd.Timestamp(self)
        return (dtg - dtg.normalize()) % cycle_length == timedelta(0)

    def set_cycle_length(self, cycle_length):
        """Set the assimilation cycle length associated with the DTG.

        Args:
            cycle_length (str): The new cycle length.

        Raises:
            ValueError: If cycle_length smaller that min allowed, or if Dtg
                not compatible with cycle_length.

        """
        # Set a default of 3H for the length of a cycle.
        # Without the context provided by a cycle length, a
        # DTG has no meaning.
        cycle_length = to_offset(cycle_length)
        # Validate cycle_length
        min_cycle_length = to_offset("1H")
        ref_date = datetime(2000, 1, 1)
        new_date = ref_date + cycle_length
        min_allowed_new_date = ref_date + min_cycle_length
        if new_date < min_allowed_new_date:
            raise ValueError(
                "Min allowed cycle_length is "
                + str(min_cycle_length)
                + ". Passed cycle_length={}".format(cycle_length)
            )
        if not self.compatible_with_cycle_length(cycle_length):
            raise ValueError(
                "Dtg {} not compatible with cycle_length {}".format(
                    self, cycle_length
                )
            )
        self._cycle_length = cycle_length

    # Override some methods to ensure results are consistent with Dtg type
    def replace(self, *args, **kwargs):
        """Overrride datetime's "replace" method.

        Args:
            *args: Args passed to datetime's original method.
            **kwargs: Keyword args passed to datetime's original method.

        Returns:
            Dtg:  Dtg with the appropriate fields replaced.

        """
        new = super().replace(*args, **kwargs)
        return Dtg(new, cycle_length=self.cycle_length)

    # Behaviour under action of + and - operands
    def __add__(self, delta):
        new_datetime = self.as_datetime() + delta
        return self.__class__(new_datetime, cycle_length=self.cycle_length)

    def __radd__(self, delta):
        return self.__add__(delta)

    def __sub__(self, other):
        new = self.as_datetime() - other
        try:
            # Casting to Dtg will work provided that "other" is a
            # timedelta-like obj and it does not violate cycle_length
            return self.__class__(new, cycle_length=self.cycle_length)
        except TypeError:
            # This will happen if "other" is any datetime-like object
            return new

    # Make sure we consider cycle_length when comparing for equality
    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        eq_timestamp = super().__eq__(other)
        same_cycle_length = self.cycle_length == other.cycle_length
        return eq_timestamp and same_cycle_length

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return self.strftime("%Y-%m-%dT%H %Z")

    def __repr__(self):
        return "{}('{}', cycle_length='{}')".format(
            self.__class__.__name__,
            self.isoformat(),
            self.cycle_length.freqstr,
        )


class DtgContainer:
    """Implement a container for instances of the Dtg class."""

    def __init__(self, data=None, start=None, end=None, cycle_length=None):
        """Validate input and set corresponding attribiutes.

        Args:
            data (Dtg or list of Dtg): Dtg or list of Dtgs.
            start (Dtg): First Dtg.
            end (Dtg): Last Dtg.
            cycle_length (str): Length of the assimilation window.

        Raises:
            ValueError: If dtg parameters are missing or inconsistent.

        """
        # Validate input:
        # (i) data and (start, end) are mutually exclusive (avoid ambiguities)
        if data is None:
            if (start is None) or (end is None):
                raise ValueError(
                    "Need 'start' and 'end' and if 'data' is not passed"
                )
            # msg_which_input and input_vals will be used later on in the
            # cycle_length validation
            msg_which_input = "'start' and 'end'"
            input_vals = [start, end]
        else:
            if (start is not None) or (end is not None):
                raise ValueError(
                    "Pass either just 'data' or both 'start' and 'end'"
                )
            msg_which_input = "the elements of 'data'"
            input_vals = data
        # (ii) Decide whether to get cycle_length from
        #      data/start/end or from class arg
        try:
            # Calling method ".freqstr" because, since pandas v1.1.0, "freq"
            # objects stopped being hashable and the following line would cause
            # TypeError "unhashable type: 'pandas._libs.tslibs.offsets.Hour'"
            data_c_length = list({d.cycle_length.freqstr for d in input_vals})
        except AttributeError:
            data_c_length = []

        if cycle_length is None:
            if len(data_c_length) == 1:
                cycle_length = data_c_length[0]
            else:
                raise ValueError(
                    "No (or inconsistent) 'cycle_length' info found in "
                    + "{}, and no 'cycle_length' ".format(msg_which_input)
                    + "passed to '{}'".format(self.__class__.__name__)
                )
        elif len(data_c_length) > 0:
            logger.warning(
                (
                    "The value(s) of the 'cycle_length' attribute(s) of "
                    "%s will be replaced by the 'cycle_length' value "
                    "passed to the '%s' constructor (=%s)"
                ),
                msg_which_input,
                self.__class__.__name__,
                cycle_length,
            )

        # Populate attributes
        self._cycle_length = to_offset(cycle_length)
        if data is None:
            self._start = Dtg(start, cycle_length=self.cycle_length)
            self._end = Dtg(end, cycle_length=self.cycle_length)
            self._data = None
        else:
            self._start = None
            self._end = None
            # Convert data to Dtg for consistency, and make it immutable
            self._data = tuple(
                Dtg(d, cycle_length=self.cycle_length) for d in data
            )

    @property
    def cycle_length(self):
        """Return the assimilation cycle length associated with the DTG."""
        return self._cycle_length

    def __len__(self):
        if self._data is None:
            max_index = abs((self._end - self._start).total_seconds())
            max_index /= self.cycle_length.nanos * 1e-9
            return int(max_index) + 1
        return len(self._data)

    def __getitem__(self, item):
        # We don't need to define __iter__ when we make an obj indexable
        if self._data is not None:
            return self._data[item]

        sign = np.sign((self._end - self._start).total_seconds())

        def calc_nth_item(n):
            if (n > len(self) - 1) or (n < -len(self)):
                raise IndexError(
                    "{} index out of range".format(self.__class__.__name__)
                )
            return self._start + sign * (n % len(self)) * self.cycle_length

        if isinstance(item, slice):
            return [calc_nth_item(i) for i in range(len(self))[item]]
        return calc_nth_item(item)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return (
            other.cycle_length == self.cycle_length
            and other._data == self._data
            and other._start == self._start
            and other._end == self._end
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        rtn = "{} object:\n".format(self.__class__.__name__)
        if self._data is None:
            rtn += "    start={}, end={}\n".format(self._start, self._end)
        else:
            rtn += "    data=[{}]\n".format(", ".join(map(str, self._data)))
        rtn += "    cycle_length={}".format(self.cycle_length)
        return rtn

    __str__ = __repr__



#!/usr/bin/env python3
"""Implement the interactive clustering app."""
import logging
import os
import tempfile
import time
from datetime import datetime

import dash
import dash_core_components as dcc
import dash_html_components as html
import dash_table
import numpy as np
import pandas as pd
import redis
from dash.dependencies import Input, Output, State
from flask_caching import Cache
from server import server

from netatmoqc.clustering import cluster_netatmo_obs, sort_df_by_cluster_size
from netatmoqc.config_parser import (
    ParsedConfig,
    UndefinedConfigValue,
    read_config,
)
from netatmoqc.domains import Domain
from netatmoqc.dtgs import Dtg
from netatmoqc.load_data import (
    read_netatmo_data_for_dtg,
    remove_irregular_stations,
)
from netatmoqc.logs import CustomFormatter
from netatmoqc.metrics import haversine_distance
from netatmoqc.plots import make_clustering_fig

logger = logging.getLogger(__name__)
logger_handler = logging.StreamHandler()
logger_handler.setFormatter(CustomFormatter())
logging.basicConfig(level=logging.INFO, handlers=[logger_handler])


config = read_config(os.getenv("NETATMOQC_CONFIG_PATH", "config.toml"))
domain = Domain.construct_from_dict(config.domain)

app = dash.Dash(
    name="clustering",
    server=server,
    url_base_pathname="/clustering/",
    meta_tags=[
        {"name": "viewport", "content": "width=device-width, initial-scale=1"}
    ],
)

# Fix duplicate log items
# TODO: Find a better way to do this without affecting Flask's own logging
if app.logger.hasHandlers():
    app.logger.handlers.clear()

app.title = "iObs NetAtmo Clustering"

# Using cache in order to share data between callbacks
# See <https://dash.plotly.com/sharing-data-between-callbacks> for details
try:
    # Test if we have a working redis server running.
    redis_server = redis.Redis("localhost", socket_connect_timeout=1)
    redis_server.ping()
    logger.info("Using redis caching")
    redis_url = os.environ.get("REDIS_URL", "redis://localhost:6379")
    CACHE_CONFIG = {"CACHE_TYPE": "redis", "CACHE_REDIS_URL": redis_url}
except redis.exceptions.ConnectionError:
    cache_dir = tempfile.TemporaryDirectory(prefix="netatmo_clustering_tmp")
    logger.info("Caching via redis unavailable. Caching to %s", cache_dir.name)
    # See <https://stackoverflow.com/questions/12868222/
    #      performance-of-redis-vs-disk-in-caching-application>
    CACHE_CONFIG = {
        "CACHE_TYPE": "filesystem",
        "CACHE_DIR": cache_dir.name,
    }
cache = Cache()
cache.init_app(app.server, config=CACHE_CONFIG)
# Clear cache on startup to avoid susprises after, e.g., updates
cache.clear()


def description_card():
    """Return a Div containing dashboard title & descriptions."""
    return html.Div(
        id="description-card",
        children=[
            html.H5("NetAtmoQC"),
            html.H3("Interactive Data Clustering"),
            html.Div(
                id="intro",
                children="Quality-Control of NetAtmo Station Data via Clustering",
            ),
        ],
        style={"text-align": "center"},
    )


def generate_obs_weights_panel():
    """Generate the observation weights panel."""

    def obs_weight_cell(var_name, default=1.0, minval=0.0, maxval=np.inf):
        """Return a div for an obs weight cell."""
        # Get defaults from config file if defined. Use the ones
        # defined in the calls to this function otherwise.
        try:
            default_from_config = config.get_clustering_opt("obs_weights", {})[
                var_name
            ]
            if default_from_config is UndefinedConfigValue:
                raise AttributeError(
                    'obs_weights not defined for "{}"'.format(var_name)
                )
            default = default_from_config
            logger.debug(
                'Using config file default "%s" for "%s" weight',
                default,
                var_name,
            )
        except (AttributeError, KeyError):
            logger.debug(
                'Using hard-coded default "%s" for "%s" weight',
                default,
                var_name,
            )

        cell = html.Div(
            id="{}_weight_div".format(var_name),
            children=[
                html.P(var_name, id="{}_weight_label".format(var_name)),
                dcc.Input(
                    id="{}_weight".format(var_name),
                    type="number",
                    inputMode="numeric",
                    min=minval,
                    max=maxval,
                    value=default,
                    style=dict(width="90%"),
                ),
            ],
            style=dict(display="table-cell"),
        )
        return cell

    panel = html.Div(
        id="obs_weights_div",
        children=[
            html.Details(
                title="Weights for the calculated differences",
                children=[
                    html.Summary("Weights for the calculated differences"),
                    html.Div(
                        children=[
                            obs_weight_cell("geo_dist", default=1.0),
                            obs_weight_cell("alt"),
                            obs_weight_cell("temperature", default=5.0),
                            obs_weight_cell("mslp"),
                        ],
                        style=dict(display="table-row"),
                    ),
                    html.Br(),
                    html.Div(
                        children=[
                            obs_weight_cell("pressure", default=0.0),
                            obs_weight_cell("humidity"),
                            obs_weight_cell("sum_rain_1"),
                        ],
                        style=dict(display="table-row"),
                    ),
                ],
                style=dict(display="table"),
            ),
        ],
    )
    return panel


allowed_cluster_methods = ["hdbscan", "dbscan", "rsl", "optics"]
allowed_outlier_rm_method = ["Iterative", "GLOSH", "LOF", "Reclustering"]


def generate_control_card():
    """Return a Div containing controls for graphs."""
    return html.Div(
        id="control-card",
        children=[
            html.P("Clustering method"),
            dcc.Dropdown(
                id="method-select",
                options=[
                    {"label": i, "value": i} for i in allowed_cluster_methods
                ],
                value=allowed_cluster_methods[0],
            ),
            html.Br(),
            # Put min_samples and eps input in their own Divs to be able to
            # put them side-by-side.
            # Adapted from <https://community.plotly.com/t/setting-the-layout-
            #               of-a-button-and-input-box/6519/3>
            html.Div(
                [
                    html.P("min_samples"),
                    dcc.Input(
                        id="min_samples",
                        type="number",
                        inputMode="numeric",
                        min=1,
                        value=5,
                        step=1,
                        required=True,
                        style=dict(display="table-cell",),
                    ),
                ],
                style=dict(display="table-cell"),
            ),
            # The min_cluster_size input should not show when using dbscan.
            # The way to create conditional inputs in dash is to put them in
            # separate Divs and change the Divs' "display" attr to "none" in
            # a callback.
            # Adapted from: <https://stackoverflow.com/questions/50213761/
            #                changing-visibility-of-a-dash-component-by-
            #                updating-other-component>
            html.Div(
                id="min_cluster_size_div",
                children=[
                    html.P("min_cluster_size", id="min_cluster_size_label"),
                    dcc.Input(
                        id="min_cluster_size",
                        type="number",
                        inputMode="numeric",
                        min=2,
                        value=5,
                        step=1,
                        style=dict(display="table-cell",),
                    ),
                ],
                style=dict(display="table-cell"),
            ),
            # The eps input should not show when using hdbscan. Doing similarly
            # to min_cluster_size above.
            html.Div(
                id="eps_div",
                children=[
                    html.P("eps", id="eps_label"),
                    dcc.Input(
                        id="eps",
                        type="number",
                        inputMode="numeric",
                        min=0.0,
                        value=10.0,
                        style=dict(display="table-cell",),
                    ),
                ],
                style=dict(display="table-cell"),
            ),
            html.Br(),
            #
            html.Div(
                id="optionals_div",
                children=[
                    html.Div(
                        id="outlier_rm_method_div",
                        children=[
                            html.Br(),
                            html.P(
                                "Post-Clustering Outlier Removal (Optional)"
                            ),
                            dcc.Dropdown(
                                id="outlier_rm_method",
                                options=[{"label": "None", "value": None}]
                                + [
                                    {"label": i, "value": i.lower()}
                                    for i in allowed_outlier_rm_method
                                ],
                                value="glosh",
                            ),
                        ],
                        style={"display": "block", "text-align": "center"},
                    ),
                    html.Div(
                        # The 'style' property of this div will be
                        # set via callback
                        id="max_num_refining_iter_div",
                        children=[
                            html.Div(
                                children=[
                                    html.Br(),
                                    html.P("Max #refining iterations"),
                                    dcc.Input(
                                        id="max_num_refine_iter",
                                        type="number",
                                        inputMode="numeric",
                                        min=1,
                                        value=100,
                                        step=1,
                                    ),
                                ],
                                style={
                                    "display": "table-cell",
                                    "text-align": "center",
                                },
                            ),
                            html.Div(
                                children=[
                                    html.Br(),
                                    html.P("Max #stdev around mean"),
                                    dcc.Input(
                                        id="max_n_std_around_mean",
                                        type="number",
                                        inputMode="numeric",
                                        min=1,
                                        value=2,
                                    ),
                                ],
                                style={
                                    "display": "table-cell",
                                    "text-align": "center",
                                },
                            ),
                        ],
                        style=dict(display="table-row"),
                    ),
                ],
            ),
            html.Br(),
            #
            generate_obs_weights_panel(),
            html.Br(),
            #
            html.Div(
                id="dtg_div",
                children=[
                    html.Div(
                        id="date-div",
                        children=[
                            html.P("Date"),
                            dcc.DatePickerSingle(
                                id="date-picker-select",
                                min_date_allowed=datetime(2018, 4, 1),
                                max_date_allowed=datetime.today(),
                                initial_visible_month=datetime(2018, 4, 1),
                                date="2018-04-01",
                                display_format="YYYY-MM-DD",
                            ),
                        ],
                        style=dict(display="table-cell"),
                    ),
                    html.Div(
                        id="cycle-div",
                        children=[
                            html.P("Cycle"),
                            dcc.Input(
                                id="cycle-select",
                                type="number",
                                inputMode="numeric",
                                min=0,
                                max=21,
                                value=0,
                                step=3,
                            ),
                        ],
                        style=dict(display="table-cell"),
                    ),
                ],
                style=dict(display="table-row"),
            ),
            #
            html.Br(),
            html.Div(
                id="plot-btn-outer",
                children=[
                    html.Button(
                        id="plot-btn",
                        children="Run Clustering and Plot",
                        n_clicks=0,
                    ),
                ],
            ),
        ],
    )


# returns top indicator div
def indicator(text, id_value):
    """Return the html.Div for one app indicator."""
    # Adapted from Dash gallery's app "dash-salesforce-crm"
    return html.Div(
        id="{}_div".format(id_value),
        children=[
            html.Div(
                children=[
                    html.P(id=id_value, className="indicator_value",),
                    html.P(text, className="indicator_text",),
                ],
            ),
        ],
        className="indicator pretty_container",
        style=dict(textAlign="center"),
    )


def generate_indicators():
    """Generate the indicators used in the app."""
    indicators = html.Div(
        id="indicators_div",
        children=[
            indicator("Mean Silhouette Score", "silhouette_coeff_indicator"),
            indicator("# Clusters", "nclusters_indicator"),
            indicator("# Accepted Obs", "naccepted_indicator"),
            indicator("# Clustering Noise", "nrejected_indicator"),
            indicator("# Clustering Outliers", "nremoved_indicator"),
            indicator("# Total Rejected", "nremoved_total_indicator"),
        ],
        style=dict(display="flex"),
    )
    return indicators


def generate_right_column_elements():
    """Generate the elements of the right-hand side column."""
    children = [
        html.Div(
            id="clustering_plot_div",
            children=[dcc.Graph(id="clustering_plot",)],
        ),
        html.Div(
            id="clustered_data_table_card",
            children=[
                html.B("Data after processing by clustering algorithm"),
                html.Hr(),
                dash_table.DataTable(
                    id="clustered_data_table",
                    data=[],
                    columns=[],
                    export_columns="all",
                    sort_action="native",
                    filter_action="native",
                    # Styling
                    style_cell=dict(padding="10px",),
                    style_header=dict(
                        backgroundColor="rgb(2,21,70)",
                        color="white",
                        textAlign="center",
                    ),
                    # Control table scrolling
                    # Don't use fixed_rows right now. It causes formatting
                    # issues at the moment (tested with dash v1.11 and v1.12)
                    style_table=dict(
                        maxHeight="300px",
                        overflowY="scroll",
                        overflowX="auto",
                    ),
                ),
            ],
        ),
    ]
    return children


app.layout = html.Div(
    id="app-container",
    children=[
        # Left column
        html.Div(
            id="left-column",
            className="three columns",
            children=[description_card(), generate_control_card()]
            + [
                html.Div(
                    ["initial child"],
                    id="output-clientside",
                    style={"display": "none"},
                )
            ],
        ),
        # Right column
        html.Div(
            # Setting "style=dict(position='relative')" here so that the
            # position-related style options in the dcc.Loading work relative
            # to this container div.
            id="right-column",
            className="nine columns",
            style=dict(position="relative"),
            children=[
                generate_indicators(),
                html.B("Visualisation of Clusters"),
                html.Div(id="calculated_dist", children=[],),
                html.Hr(),
                dcc.Loading(
                    # Embed the the right-hand side column inside a dcc.Loading
                    # so that a spinner is shown while updating.
                    id="clustering_plot-spinner",
                    # type is one of:
                    #     'graph', 'cube', 'circle', 'dot', or 'default'
                    type="graph",
                    fullscreen=False,
                    children=generate_right_column_elements(),
                    # Put spinner in the middle of the container Div
                    style={
                        "position": "absolute",
                        "top": "50%",
                        "left": "50%",
                        "transform": "translate(-50%, -50%)",
                    },
                ),
            ],
        ),
    ],
)


# perform expensive computations in this "global store"
# these computations are cached in a globally available
# redis memory store which is available across processes
# and for all time.
@cache.memoize()
def read_data_df(str_date, cycle):
    """Read data for DTG composed by str_date and cycle."""
    dtg = Dtg(datetime.strptime(str_date, "%Y-%m-%d").replace(hour=cycle))
    df = read_netatmo_data_for_dtg(dtg, rootdir=config.general.data_rootdir)
    df, _ = remove_irregular_stations(df)
    return df


# Show/hide eps input when appropriate.
@app.callback(
    [
        Output(component_id="eps_div", component_property="style"),
        Output(
            component_id="min_cluster_size_div", component_property="style"
        ),
    ],
    [Input(component_id="method-select", component_property="value")],
)
def show_hide_depending_on_method(method):
    """Show/hide menus depending on clustering method."""
    if method == "dbscan":
        eps_div_style = {"display": "table-cell"}
        min_cluster_size_div_style = {"display": "none"}
    else:
        eps_div_style = {"display": "none"}
        min_cluster_size_div_style = {"display": "table-cell"}
    return (eps_div_style, min_cluster_size_div_style)


#
@app.callback(
    [
        Output(
            component_id="max_num_refining_iter_div",
            component_property="style",
        ),
    ],
    [Input(component_id="outlier_rm_method", component_property="value")],
)
def show_hide_max_num_refining_iter(outlier_rm_method):
    """Show/hide max_num_refining_iter menu according to outlier_rm_method."""
    if outlier_rm_method == "iterative":
        return [{"display": "table-row"}]
    return [{"display": "none"}]


# Producing plot
@app.callback(
    [
        Output("clustering_plot", "figure"),
        Output("clustered_data_table", "columns"),
        Output("clustered_data_table", "data"),
        Output("silhouette_coeff_indicator", "children"),
        Output("nclusters_indicator", "children"),
        Output("naccepted_indicator", "children"),
        Output("nrejected_indicator", "children"),
        Output("nremoved_indicator", "children"),
        Output("nremoved_total_indicator", "children"),
    ],
    [Input("plot-btn", "n_clicks")],
    [
        State("method-select", "value"),
        State("min_samples", "value"),
        State("min_cluster_size", "value"),
        State("eps", "value"),
        State("date-picker-select", "date"),
        State("cycle-select", "value"),
        State("outlier_rm_method", "value"),
        State("max_num_refine_iter", "value"),
        State("max_n_std_around_mean", "value"),
        State("geo_dist_weight", "value"),
        State("alt_weight", "value"),
        State("temperature_weight", "value"),
        State("mslp_weight", "value"),
        State("pressure_weight", "value"),
        State("humidity_weight", "value"),
        State("sum_rain_1_weight", "value"),
    ],
)
def run_clustering_and_make_plot(
    n_clicks,
    method,
    min_samples,
    min_cluster_size,
    eps,
    date,
    cycle,
    outlier_rm_method,
    max_num_refining_iter,
    refine_max_std,
    geo_dist_weight,
    alt_weight,
    temperature_weight,
    mslp_weight,
    pressure_weight,
    humidity_weight,
    sum_rain_1_weight,
):
    """Control the app. Main app routine."""
    # pylint: disable=locally-disabled, too-many-locals, too-many-arguments
    empty_fig = make_clustering_fig(pd.DataFrame(), domain=domain)
    empty_rtn = (empty_fig, [], [], "-", "-", "-", "-", "-", "-")
    if n_clicks == 0:
        return empty_rtn

    logger.info("Reading data...")
    start_read_data = time.time()
    df = read_data_df(date, cycle)
    end_read_data = time.time()
    logger.info(
        "Done reading data. Elapsed: %.1fs", end_read_data - start_read_data
    )

    n_obs = len(df.index)
    if n_obs == 0:
        logger.error("ERROR:  No data found for selected DTG.")
        return empty_rtn

    # Create a config obj with params
    clustering_config = ParsedConfig(
        {
            "general": dict(
                clustering_method=method,
                custom_metrics_optimize_mode=config.general.custom_metrics_optimize_mode,
                dtgs=dict(
                    list=config.general.dtgs,
                    cycle_length=config.general.dtgs.cycle_length.freqstr,
                ),
            ),
            "clustering_method.%s"
            % (method): dict(
                eps=eps,
                min_cluster_size=min_cluster_size,
                min_samples=min_samples,
                outlier_removal={
                    "method": outlier_rm_method,
                    outlier_rm_method: dict(
                        max_n_iter=max_num_refining_iter,
                        max_n_stdev=abs(refine_max_std),
                    ),
                },
                obs_weights=dict(
                    geo_dist=geo_dist_weight,
                    alt=alt_weight,
                    temperature=temperature_weight,
                    mslp=mslp_weight,
                    pressure=pressure_weight,
                    humidity=humidity_weight,
                    sum_rain_1=sum_rain_1_weight,
                ),
            ),
            "domain": config.domain,
        }
    )

    time_start_clustering = time.time()
    logger.info("Running %s...", method)
    df = cluster_netatmo_obs(
        df=df, config=clustering_config, calc_silhouette_samples=True,
    )
    df = sort_df_by_cluster_size(df)
    time_end_clustering = time.time()
    logger.info(
        "%s performed in %.1fs",
        method,
        time_end_clustering - time_start_clustering,
    )

    noise_data_df = df[df["cluster_label"] < 0]
    n_noise_clusters = len(noise_data_df["cluster_label"].unique())
    noise_count = len(noise_data_df)
    n_rm_refining = (df["cluster_label"] < -1).sum()
    n_rm_clustering = noise_count - n_rm_refining
    n_clusters = len(df["cluster_label"].unique()) - n_noise_clusters
    n_accepted = n_obs - noise_count
    silhouette_score = df["silhouette_score"].mean(skipna=True)

    logger.debug("Estimated number of clusters: %d", n_clusters)
    logger.debug("Estimated number of noise points: %d", noise_count)
    logger.debug("Estimated number of clustered obs: %d", n_accepted)
    logger.debug("Mean silhouette score: %.3f", silhouette_score)

    fig = make_clustering_fig(df, domain=domain)

    # Preparing update to table
    # Formatting column values for better presentation
    for cname in df.columns:
        if cname in ["lat", "lon"]:
            df[cname] = df[cname].map("{:.4f}".format)
        elif cname in [
            "alt",
            "pressure",
            "mslp",
            "temperature",
            "humidity",
            "sum_rain_1",
        ]:
            df[cname] = df[cname].map("{:.1f}".format)
        elif cname in ["silhouette_score", "GLOSH", "LOF"]:
            df[cname] = df[cname].map("{:.2f}".format)
    df["time_utc"] = df["time_utc"].dt.strftime("%Y-%m-%d %H:%M:%S")
    # We'll set column IDs to the same values as their col names, but
    # this is not a requirement
    columns = [{"name": n, "id": n, "selectable": True} for n in df.columns]
    data = df.to_dict("records")

    return (
        fig,
        columns,
        data,
        dcc.Markdown("**{:.2f}**".format(silhouette_score)),
        dcc.Markdown("**{}**".format(n_clusters)),
        dcc.Markdown(
            "**{} ({:.2f}%)**".format(n_accepted, 100.0 * n_accepted / n_obs)
        ),
        dcc.Markdown(
            "**{} ({:.2f}%)**".format(
                n_rm_clustering, 100.0 * n_rm_clustering / n_obs
            )
        ),
        dcc.Markdown(
            "**{} ({:.2f}%)**".format(
                n_rm_refining, 100.0 * n_rm_refining / n_obs
            )
        ),
        dcc.Markdown(
            "**{} ({:.2f}%)**".format(noise_count, 100.0 * noise_count / n_obs)
        ),
    )


# Report distance between points upon selection fo two points
@app.callback(
    [Output(component_id="calculated_dist", component_property="children")],
    [
        # Set clickmode='event+select' in the figure layout, and then
        # use 'selectedData' here instead of 'clickData'
        Input(
            component_id="clustering_plot", component_property="selectedData"
        ),
    ],
)
def geodist_upon_pt_pair_selection(selected_data):
    """Return geodist between selected points in the app."""
    if (selected_data is None) or (len(selected_data["points"]) != 2):
        return [html.P("")]

    pt0 = [selected_data["points"][0][attr] for attr in ["lat", "lon"]]
    pt1 = [selected_data["points"][1][attr] for attr in ["lat", "lon"]]
    rtn = html.Span(
        "Haversine distance between selected points: {:.3f} km".format(
            haversine_distance(np.array(pt0), np.array(pt1))
        )
    )
    return [rtn]


# Run the server
if __name__ == "__main__":
    app.run_server(debug=True)



#!/usr/bin/env python3
"""Implement the package's commands."""
import logging
import multiprocessing
import os
import platform
import subprocess
import time
from datetime import datetime
from functools import partial

import humanize
import numpy as np
import pandas as pd
import psutil
from joblib import Parallel, delayed, parallel_backend

from .clustering import (
    cluster_netatmo_obs,
    report_clustering_results,
    sort_df_by_cluster_size,
)
from .config_parser import read_config
from .domains import Domain
from .dtgs import Dtg
from .load_data import (
    DataNotFoundError,
    read_netatmo_data_for_dtg,
    remove_irregular_stations,
    rm_moving_stations,
)
from .logs import get_logger, logcolor
from .mpi import mpi_parallel
from .plots import (
    DEF_FIGSHOW_CONFIG,
    make_clustering_fig,
    show_cmd_get_fig_from_dataframes,
)
from .save_data import netatmoqc_input2output

logger = logging.getLogger(__name__)


def cluster_obs_single_dtg(args):
    """Implement the "cluster" command.

    Args:
        args (argparse.Namespace): Parsed command line arguments.

    """
    config = read_config(args.config_file)

    try:
        dtg = Dtg(args.dtg)
    except TypeError:
        dtg = None
    if dtg is None:
        dtg = config.general.dtgs[0]
    logger.info(
        "Running %s on obs for %sDTG=%s%s...",
        config.general.clustering_method,
        logcolor.cyan,
        dtg,
        logcolor.reset,
    )

    if args.savefig:
        # Create outdir at the beginning so users don't
        # waste time in case they can't save results
        outdir = config.general.outdir / "{}_netatmoqc_cluster".format(
            datetime.now().strftime("%Y-%m-%d_%H.%M.%S")
        )
        # Allow mkdir to raise eventual exceptions if cannot write to outdir
        outdir.mkdir(parents=True)

    try:
        df = read_netatmo_data_for_dtg(
            dtg, rootdir=config.general.data_rootdir
        )
    except DataNotFoundError:
        logger.warning(
            "Could not cluster obs for dtg=%s: ", dtg, exc_info=True
        )
        return

    df, _ = remove_irregular_stations(df)
    df = cluster_netatmo_obs(df=df, config=config)
    report_clustering_results(df)

    if args.show or args.savefig:
        logger.info("Preparing figure...")
        df = sort_df_by_cluster_size(df)
        domain = Domain.construct_from_dict(config.domain)
        fig = make_clustering_fig(df, domain=domain)
        if args.savefig:
            fpath = str(outdir / "cluster_obs_single_dtg.html")
            logger.info("Saving figure to %s\n", fpath)
            fig.write_html(fpath)
        if args.show:
            fig.show(config=DEF_FIGSHOW_CONFIG)

    logger.info(
        "%sDone with 'cluster' command.%s", logcolor.cyan, logcolor.reset
    )


def _select_stations_single_dtg(dtg, config, args):
    """Implement "select_stations" for a single DTG.

    Args:
        dtg (Dtg): The dtg for which to select stations.
        config (ParsedConfig): Parsed dict of configs from config file.
        args (argparse.Namespace): Parsed command line arguments.

    Returns:
        (DataFrame, DataFrame, DataFrame): df_accepted, df_rejected, df_moving

    """
    tstart = time.time()
    logger = get_logger(__name__, args.loglevel)
    logger.info("%sDTG=%s%s: Started", logcolor.cyan, dtg, logcolor.reset)

    # Some control of oversubscription if using mpi.
    # This is not needed when using single-host joblib with "loky"
    cpu_share = -1
    if args.mpi:
        proc_parent = psutil.Process(os.getppid())
        proc_family_size = len(proc_parent.children())
        cpu_share = multiprocessing.cpu_count() // proc_family_size

    try:
        df = read_netatmo_data_for_dtg(
            dtg, rootdir=config.general.data_rootdir
        )
    except DataNotFoundError:
        logger.warning("Could not select obs for dtg=%s: ", dtg, exc_info=True)
        return pd.DataFrame(), pd.DataFrame(), pd.DataFrame()

    df, df_moving = remove_irregular_stations(df)

    df = cluster_netatmo_obs(
        df=df, config=config, n_jobs=cpu_share, calc_silhouette_samples=False,
    )

    selected_cols = ["id", "lat", "lon", "alt"]
    rejected_stat_mask = df["cluster_label"] < 0
    df_rejected = df[selected_cols][rejected_stat_mask]
    df_accepted = df[selected_cols][~rejected_stat_mask]
    df_moving = df_moving[selected_cols]

    logger.info("DTG=%s: Done. Elapsed: %.1fs", dtg, time.time() - tstart)
    return df_accepted, df_rejected, df_moving


def select_stations(args):
    """Implement the "select" command.

    This function calls "_select_stations_single_dtg" for each DTG
    (in parallel if requested/possible), and then processes, gathers
    and saves the results.

    Args:
        args (argparse.Namespace): Parsed command line arguments.

    """
    tstart_selection = time.time()
    config = read_config(args.config_file)

    # Create outdir at the beginning so users don't
    # waste time in case they can't save results
    outdir = config.general.outdir / "{}_netatmoqc_select".format(
        datetime.now().strftime("%Y-%m-%d_%H.%M.%S")
    )
    # Allow mkdir to raise eventual exceptions if cannot write to outdir
    outdir.mkdir(parents=True)

    # Process DTGs in parallel if possible/requested
    if args.mpi:
        logger.info("Parallelising tasks over DTGs using MPI")
        func = partial(_select_stations_single_dtg, config=config, args=args)
        results = mpi_parallel(func, config.general.dtgs)
    else:
        n_procs = int(os.getenv("NETATMOQC_MAX_PYTHON_PROCS", "1"))
        if n_procs > 1:
            logger.info("Parallelising tasks over DTGs within a single host")
        # Using the "loky" backend helps avoiding oversubscription
        # of cpus in child processes
        with parallel_backend("loky"):
            results = Parallel(n_jobs=n_procs)(
                delayed(_select_stations_single_dtg)(dtg, config, args)
                for dtg in config.general.dtgs
            )
    logger.info(
        "End of survey over %d DTGs. Combining results...\n",
        len(config.general.dtgs),
    )
    # Gather results: Only accepted and rejected stations for now. Moving
    # stations will be handled after post-processing accepted and rejected
    df_accepted = pd.concat((r[0] for r in results), ignore_index=True)
    df_rejected = pd.concat((r[1] for r in results), ignore_index=True)

    df_accepted, df_moving_accepted = rm_moving_stations(df_accepted)
    if len(df_moving_accepted.index) > 0:
        logger.warning(
            'Removed %d "moving" stations from the list of accepted stations',
            len(df_moving_accepted["id"].unique()),
        )

    df_rejected, df_moving_rejected = rm_moving_stations(df_rejected)
    if len(df_moving_rejected.index) > 0:
        logger.warning(
            'Removed %d "moving" stations from the list of removed stations',
            len(df_moving_rejected["id"].unique()),
        )

    # Let's recover some stations for the list of rejected if they
    # also appear among the df_accepted. Do this based on a
    # tolerance criteria.
    logger.info(
        (
            "Applying station_rejection_tol=%.2f to rescue "
            'stations from the "rejected" list'
        ),
        config.commands.select.station_rejection_tol,
    )
    accept_counts_per_id = df_accepted["id"].value_counts()
    reject_counts_per_id = df_rejected["id"].value_counts()
    rescued_rejected_stats = []
    # Keep track of rejection rates. We'll use this later.
    id2rejection_rate = {}
    for stat_id in reject_counts_per_id.index:
        reject_count = reject_counts_per_id[stat_id]
        try:
            accept_count = accept_counts_per_id[stat_id]
        except KeyError:
            accept_count = 0
        rejection_rate = 1.0 * reject_count / (reject_count + accept_count)
        id2rejection_rate[stat_id] = rejection_rate
        if rejection_rate < config.commands.select.station_rejection_tol:
            rescued_rejected_stats.append(stat_id)
    # Using a rescued_rejected_stats list to modify the dataframes
    # df_accepted/df_rejected is much faster than modifying
    # the dataframes inside the loop above
    rescued_rows = df_rejected.loc[
        df_rejected["id"].isin(rescued_rejected_stats), :
    ]
    df_accepted = df_accepted.append(rescued_rows, ignore_index=True)
    df_rejected = df_rejected.drop(rescued_rows.index)
    if len(rescued_rejected_stats) > 0:
        logger.info(
            (
                "    > Rescuing %d stations rejected in "
                "less than %.1f%% of occurences"
            ),
            len(rescued_rejected_stats),
            100 * config.commands.select.station_rejection_tol,
        )
    else:
        logger.info('    > No stations recovered from "rejected"')

    # We don't need duplicate entries
    df_accepted = df_accepted.drop_duplicates(subset=["id"])
    df_rejected = df_rejected.drop_duplicates(subset=["id"])

    # Now we can handle moving stations
    df_moving = pd.concat(
        [r[2] for r in results] + [df_moving_accepted, df_moving_rejected],
        ignore_index=True,
    )
    df_unique_moving = df_moving["id"].unique()

    df_rejected = df_rejected[~df_rejected["id"].isin(df_unique_moving)]
    df_accepted = df_accepted[
        ~(
            df_accepted["id"].isin(df_unique_moving)
            | df_accepted["id"].isin(df_rejected["id"])
        )
    ]
    if len(df_unique_moving) > 0:
        logger.warning(
            (
                "\n A total of %d unique stations were removed from both\n"
                " accepted and rejected lists because they changed\n"
                " at least one of (lat, lon, alt) throughout the survey\n"
                " These will be saved separately in the moving_stations.csv file"
            ),
            len(df_unique_moving),
        )

    if len(df_accepted.index) == 0:
        logger.info(
            "%sSelection finished. No stations were accepted. "
            "%d stations were rejected.%s",
            logcolor.cyan,
            len(df_rejected.index),
            logcolor.reset,
        )
        return

    ##########################################################################
    # Trim accepted obs to (a possibly coarser version of) the domain's grid #
    ##########################################################################
    # Add (rejection_rate, grid_i, grid_j) columns to df_accepted,           #
    # then, for each (grid_i, grid_j), pick only the station that has the    #
    # lowest rejection_rate                                                  #
    ##########################################################################
    coarse_grid_factor = config.domain.thinning_grid_coarse_factor
    if coarse_grid_factor > 0:
        domain = Domain.construct_from_dict(config.domain)
        logger.info(
            "Thinning accepted obs: Keep only 1 station per support grid point"
        )
        logger.info(
            "    > Support grid spacing: %.1f m (%d x the domain's)",
            domain.thinning_grid.x_spacing,
            coarse_grid_factor,
        )
        n_preliminary_accepted = len(df_accepted.index)

        # (a) Add rejection rate info
        def rej_rate_non_vec(stat_id):
            try:
                return id2rejection_rate[stat_id]
            except KeyError:
                return 0.0

        rej_rate = np.vectorize(rej_rate_non_vec, otypes=[float])

        df_accepted["rejection_rate"] = rej_rate(df_accepted["id"])

        # (b) Sort by rejection rate (lower to higher)
        df_accepted = df_accepted.loc[
            df_accepted["rejection_rate"].sort_values(ascending=True).index
        ]

        # (c) Add grid (i, j) info
        icol, jcol = domain.thinning_grid.lonlat2grid(
            df_accepted["lon"].to_numpy(), df_accepted["lat"].to_numpy(),
        )
        df_accepted["i"] = icol
        df_accepted["j"] = jcol

        # (d) rm all but the lowest-rejection-rate station at each grid (i, j)
        grid_trimmed_stations = df_accepted.groupby(
            ["i", "j"], as_index=False, sort=False
        ).first()
        grid_trimmed_stations = grid_trimmed_stations.drop(["i", "j"], axis=1)
        df_accepted = df_accepted.drop(["i", "j"], axis=1)

        # (e) Finally, move to df_rejected those stations that appear in
        #     df_accepted but not in grid_trimmed_stations
        grid_trimmed_mask = df_accepted["id"].isin(grid_trimmed_stations["id"])
        df_rejected = df_rejected.append(df_accepted[~grid_trimmed_mask])
        df_accepted = grid_trimmed_stations

        logger.info(
            "    > Number of stations rejected at the thinning step: %d",
            n_preliminary_accepted - len(df_accepted.index),
        )

    ##################
    # Report results #
    ##################
    total_nstations = sum(
        map(len, [df_accepted.index, df_rejected.index, df_unique_moving])
    )
    n_selected = len(df_accepted.index)
    ratio = 100.0 * n_selected / total_nstations

    logger.info(
        "%sFinished with %d accepted stations%s (%.2f%% out of %d)",
        logcolor.cyan,
        n_selected,
        logcolor.reset,
        ratio,
        total_nstations,
    )
    logger.info(
        "That's %d rejections %s(%.2f%% rejection rate)%s",
        total_nstations - n_selected,
        logcolor.cyan,
        100.0 - ratio,
        logcolor.reset,
    )

    ###########################
    # Save and/or show results #
    ###########################
    fpath = outdir / "accepted_stations.csv"
    logger.info(
        "%sSaving accepted station data to%s %s",
        logcolor.cyan,
        logcolor.reset,
        fpath,
    )
    df_accepted.to_csv(fpath, index=None, mode="w")

    fpath = outdir / "rejected_stations.csv"
    logger.info(
        "%sSaving rejected station data to%s %s",
        logcolor.cyan,
        logcolor.reset,
        fpath,
    )
    df_rejected.to_csv(fpath, index=None, mode="w")

    fpath = outdir / "moving_stations.csv"
    logger.info(
        "%sSaving moving station data%s to %s\n",
        logcolor.cyan,
        logcolor.reset,
        fpath,
    )
    df_moving.to_csv(fpath, index=None, mode="w")

    # Save data from selected stations in OBSOUL file format
    netatmoqc_input2output(
        config.general.dtgs,
        netatmo_data_rootdir=config.general.data_rootdir,
        selected_stations=df_accepted["id"],
        outdir=outdir,
        loglevel=args.loglevel,
        use_mpi=args.mpi,
        save_csv=True,
        save_obsoul=args.save_obsoul,
        obsoul_export_params=config.general.obsoul_export_params,
    )

    # We're now done.
    elapsed = time.time() - tstart_selection
    logger.info(
        "%sDone with 'select' command, %d DTGs. Elapsed: %s (~%s per DTG)%s",
        logcolor.cyan,
        len(config.general.dtgs),
        humanize.precisedelta(elapsed),
        humanize.precisedelta(elapsed / len(config.general.dtgs)),
        logcolor.reset,
    )


# Code related to the csv2obsoul command
def csv2obsoul(args):
    """Implement the "csv2obsoul" command.

    Args:
        args (argparse.Namespace): Parsed command line arguments.

    """
    config = read_config(args.config_file)

    # Create outdir at the beginning so users don't
    # waste time in case they can't save results
    outdir = config.general.outdir / "{}_netatmoqc_csv2obsoul".format(
        datetime.now().strftime("%Y-%m-%d_%H.%M.%S")
    )
    # Allow mkdir to raise eventual exceptions if cannot write to outdir
    outdir.mkdir(parents=True)

    # Allow picking only selected stations
    selected_stations = None
    if args.selected_stations_fpath is not None:
        if args.selected_stations_fpath.suffix == ".csv":
            selected_stations = pd.read_csv(args.selected_stations_fpath)["id"]
            logger.info(
                "Found %s selected stations in file '%s'.",
                len(selected_stations),
                args.selected_stations_fpath,
            )
        else:
            raise NotImplementedError(
                'Only csv files supported in "--selected-stations-fpath". '
                "Received '%s'." % (args.selected_stations_fpath),
            )

    netatmoqc_input2output(
        config.general.dtgs,
        netatmo_data_rootdir=config.general.data_rootdir,
        dropna=args.dropna,
        fillna=args.fillna,
        selected_stations=selected_stations,
        rm_duplicate_stations=args.rm_duplicate_stations,
        rm_moving_stations=args.rm_moving_stations,
        outdir=outdir,
        loglevel=args.loglevel,
        use_mpi=args.mpi,
        save_csv=False,
        save_obsoul=True,
        obsoul_export_params=config.general.obsoul_export_params,
    )

    logger.info(
        "%sDone with 'csv2obsoul' command.%s", logcolor.cyan, logcolor.reset
    )


# Code related to the "show" command
def _open_file_with_default_app(fpath):
    if platform.system() == "Windows":
        os.startfile(fpath)
    elif platform.system() == "Darwin":
        subprocess.call(("open", fpath))
    else:
        subprocess.call(("xdg-open", fpath))


def show(args):
    """Implement the 'show' command.

    Args:
        args (argparse.Namespace): Parsed command line arguments.

    """
    logger = get_logger(__name__, args.loglevel)

    config = read_config(args.config_file)
    domain = Domain.construct_from_dict(config.domain)

    if args.domain:
        fig = domain.get_fig(display_grid_max_gsize=20000.0)
        fig.update_layout(
            legend=dict(
                orientation="h", xanchor="center", x=0.5, yanchor="top", y=0.0,
            )
        )
        fig.show(config=DEF_FIGSHOW_CONFIG)

    if args.config:
        logger.info("TOML config dump:\n\n%s", config.toml_dumps())

    dataframes = {}
    for fpath in args.file_list:
        if fpath.suffix == ".csv":
            logger.info("Reading data from file %s", fpath)
            dataframes[fpath.name] = pd.read_csv(fpath)
        elif fpath.suffix in [".htm", ".html"]:
            logger.info("Openning file '%s'", fpath)
            _open_file_with_default_app(fpath)
        else:
            logger.warning(
                "Only html and csv files supported. Skipping file '%s'", fpath
            )

    if len(dataframes) > 0:
        fig = show_cmd_get_fig_from_dataframes(args, dataframes, domain)
        fig.show(config=DEF_FIGSHOW_CONFIG)



#!/usr/bin/env python3
"""Code related to saving output files."""
import logging
import os
from collections import OrderedDict
from functools import partial
from pathlib import Path

import numpy as np
import pandas as pd
import psutil
from joblib import Parallel, delayed, parallel_backend

from .dtgs import datetime2epoch
from .load_data import (
    DataNotFoundError,
    read_netatmo_data_for_dtg,
    remove_irregular_stations,
)
from .logs import get_logger, logcolor
from .mpi import mpi_parallel

logger = logging.getLogger(__name__)


def save_df_as_netatmo_csv(df, path, overwrite=False):
    """Save dataframe in the same CSV format netatmoqc reads.

    Args:
        df (pandas.Dataframe): The dataframe to be saved.
        path (pathlib.Path): Path for the output file.
        overwrite (bool): Overwite or not the out file. Default value = False.

    Raises:
        FileExistsError: If path doesn't exist and overwrite==False.
        ValueError: If df["time_utc"] is nether an integer (assumed to be an
            epoch) nor a datetime object.

    """
    path = Path(path).resolve()
    if path.exists() and not overwrite:
        raise FileExistsError("File {} exists".format(path))
    path.parent.mkdir(parents=True, exist_ok=True)

    if pd.api.types.is_datetime64_any_dtype(df["time_utc"]):
        df["time_utc"] = datetime2epoch(df["time_utc"])
    elif not pd.api.types.is_integer_dtype(df["time_utc"]):
        raise ValueError("time_utc is nether an integer nor a datetime")

    # It hurts a bit to label pressures as mslp again, but this is needed
    # so that the csv files generated by this routine are consistent
    # in data type with the original ones provided by netatmo
    if "pressure" in df.columns and "mslp" in df.columns:
        # In this case, we have recovered pressure from netatmo original file
        # when reading the df
        df = df.drop(["pressure"], axis=1)
    df = df.rename(columns=dict(mslp="pressure"))

    # Saving the adjusted df
    logger.debug(
        "%sSaving CSV, %d obs%s: %s",
        logcolor.cyan,
        len(df.index),
        logcolor.reset,
        path,
    )
    df.to_csv(path, index=None, mode="w")


def save_df_as_obsoul(df, fpath=None, export_params=None):
    """Save dataframe returned by read_netatmo_data_for_dtg in obsoul format.

    For further reference on the obsoul file format, please take a look at the
    documentation file available at
    <http://www.rclace.eu/File/Data_Assimilation/2007/lace_obspp.pdf>

    Args:
        df (netatmoqc.load_data.DataFrameWithDtgMetadata): Input dataframe.
            This type is inherited from pandas.DataFrame, but it has added
            DTG metadata info (df.metadata_dict["dtg"]).
        fpath (pathlib.Path):  Output file path (Default value = OBSOUL${DTG}).
        export_params (list): Parameters to save in the OBSOULD file.
            (Default value=["pressure","temperature","humidity","sum_rain_1"])

    """
    if export_params is None:
        export_params = ["pressure", "temperature", "humidity", "sum_rain_1"]

    dtg = df.metadata_dict["dtg"]

    # Define n_params, the "number of bodies" in obsoul lingo
    df_data_cols = [_ for _ in export_params if _ in df.columns]
    n_params = len(df_data_cols)
    if n_params == 0:
        logger.warning("Cannot produce file '%s': No valid export_params.")
        return
    # Remove duplicates from df_data_cols. Keep order so output files
    # can be compared using "diff" (we'll iterate over df_data_cols)
    df_data_cols = list(OrderedDict.fromkeys(df_data_cols))

    # More on comparability via diff: Sort df rows as we'll iterate over them
    key_cols = ["time_utc", "lat", "lon", "alt", "id"]
    sort_by = key_cols + [_ for _ in df_data_cols if _ not in key_cols]
    df = df.sort_values(by=sort_by, ignore_index=True)

    # Validate fpath and create parent dirs tree if not in place
    if fpath is None:
        fpath = Path() / "OBSOUL{}".format(dtg.strftime("%Y%m%d%H"))
    fpath = Path(fpath)
    fpath.parent.mkdir(parents=True, exist_ok=True)

    # Convert pressure from hPa to Pa
    if "pressure" in df_data_cols:
        df["pressure"] *= 100.0
    if "mslp" in df_data_cols:
        df["mslp"] *= 100.0

    # Define some constants
    obs_type = 1  # 1: SYNOP
    # Which obs_code to use?
    #  * obs_code=14: Automated Land Synop (seems reasonable)
    #  * obs_code=199: New codetype (used as codetype in Roel's csv2obsoul.jl)
    obs_code = 199
    obs_quality_flag = "1111"  # Got from Jelena
    site_dependent_flag = 100000  # Got from Jelena
    var2varcode = dict(
        # I got these from https://apps.ecmwf.int/odbgov/varno/
        mslp=108,  # varname="pmsl"
        pressure=110,  # varname="ps"
        temperature=39,  # varname="t2m"
        humidity=58,  # varname="rh2m". Or should I use 7 (varname=q)?
        # sum_rain_1 is prob not OK. I got it from
        # https://apps.ecmwf.int/odbgov/accumulationkind/
        sum_rain_1=39001,
    )

    # Start file format conversion
    analysis_date = dtg.strftime("%Y%m%d")
    analysis_time = dtg.strftime("%H")
    with open(fpath, "w") as obsoul_file:
        # The "date time" first line in the file
        obsoul_file.write("{} {}\n".format(analysis_date, analysis_time))

        for row in df.itertuples(index=False):
            # Construct the records. Each record has a header and multiple
            # "bodies", all in the same line.
            #
            # 1) Construct header.
            # We should have one header for each (id, lat, lon, alt). Since we
            # forbid, in this package, that stations move (lat, lon, alt) in
            # any given DTG, this implies that we should have one -- and only
            # one - header for each id. Moreover, as we are also removing
            # duplicate station IDs within any given DTG, this therefore
            # implies that there should be one and only one header for each
            # row in our input dataframe.
            obs_date = row.time_utc.strftime("%Y%m%d")
            # obs_hour must not have leading zeros for the hour
            obs_hour = row.time_utc.strftime("%k%M%S").strip()
            header = (
                17,  # Got this from Jelena
                obs_type,
                obs_code,
                row.lat,
                row.lon,
                "'{}'".format(row.id),
                obs_date,
                obs_hour,
                row.alt,
                n_params,
                obs_quality_flag,
                site_dependent_flag,
            )

            # Write the header. According to the obsoul file format docs, each
            # observation record (header + bodies) should occupy only one line
            # in the file, so let's skip putting a newline between header and
            # bodies and just put a space at the beginning of each body
            obsoul_file.write(" ".join(map(str, header)))

            # Construct bodies
            for param_name in df_data_cols:
                # param_type is the same as "varid@body" in odb lingo
                param_type = var2varcode[param_name]

                # param_quality_flag:
                #   * Jelena said it could maybe be "ne"
                #   * Roel uses paramqcflag=2048 in his csv2obsoul.jl
                # I'm using the same as Roel for consistency
                param_quality_flag = 2048

                # Set first_vert_coord similarly as in  Roel's csv2obsoul.jl
                # g0: Globally-averaged gravity acceleration in m*s^-2
                g0 = 9.81
                first_vert_coord = g0 * row.alt

                # vertco_reference_2: 2nd vertical coord reference. We don't
                # have a value for that. Signal missing value with a sentinel
                # value of 0.1699999976E+39 (according to the obsoul doc file)
                vertco_reference_2 = 0.1699999976e39

                body = (
                    param_type,
                    first_vert_coord,
                    vertco_reference_2,
                    getattr(row, param_name),
                    param_quality_flag,
                )
                # Again, no space between the bodies of the same record,
                # but just a space at the beginning of each body
                obsoul_file.write(" " + " ".join(map(str, body)))

            # All bodies for this record are written. Newline is finally needed
            obsoul_file.write("\n")


@np.vectorize
def obs_timestamp2csv_rpath(timestamp):
    """Get appropriate relative CSV file path from observation timestamp.

    Args:
        timestamp (datetime.datetime): The timestamp of the observation.

    Returns:
        pathlib.Path: CSV file relative path to be used.

    """
    year = timestamp.strftime("%Y")
    month = timestamp.strftime("%m")
    day = timestamp.strftime("%d")
    minute = int(timestamp.strftime("%M"))
    if minute < 10:
        f_minute = "05"
    elif minute < 20:
        f_minute = "15"
    elif minute < 30:
        f_minute = "25"
    elif minute < 40:
        f_minute = "35"
    elif minute < 50:
        f_minute = "45"
    elif minute < 60:
        f_minute = "55"

    fname = "%s%s00Z.csv" % (timestamp.strftime("%Y%m%dT%H"), f_minute)
    return Path("%s/%s/%s/%s" % (year, month, day, fname))


def _input2output_single_dtg(
    dtg,
    netatmo_data_rootdir,
    selected_stations=None,
    dropna=True,
    fillna=None,
    rm_duplicate_stations=True,
    rm_moving_stations=True,
    outdir_csv=None,
    outdir_obsoul=None,
    obsoul_export_params=None,
    loglevel="INFO",
):
    """Read data from one DTG, filter selected obs and save results.

    Args:
        dtg (netatmoqc.dtgs.Dtg): Assimilation DTG for the data.
        netatmo_data_rootdir (pathlib.Path): Path to the top-level directory
            where the netatmo input csv files are stored.
        selected_stations (list): Selected station IDs (Default value = None).
        dropna (bool): Drop NaN values before saving (Default value = True).
        fillna (dict): Dictionary having data parameters as keys and the fill
            values as dict values. Passed to read_netatmo_csv
            (Default value = None).
        rm_duplicate_stations (bool): (Default value = True)
        rm_moving_stations (bool): (Default value = True)
        outdir_csv (pathlib.Path): Dir where out csv files will be put. If None
            is passed, then no csv files will be saved. (Default value = None)
        outdir_obsoul (pathlib.Path): Analogous to outdir_csv, but this time
            for obsoul files.
        obsoul_export_params (list): List of data params to be exported in the
            final obsoul file. Passed to save_df_as_obsoul.
            (Default value = None).
        loglevel (str): Level for logging (Default value = "INFO").

    """
    logger = get_logger(__name__, loglevel)

    logger.debug(
        "Reading data for %sDTG=%s%s...", logcolor.cyan, dtg, logcolor.reset,
    )
    try:
        # read_netatmo_data_for_dtg will raise DataNotFoundError if
        # there's no data for the selected DTG
        df = read_netatmo_data_for_dtg(
            dtg,
            rootdir=netatmo_data_rootdir,
            dropna=dropna,
            fillna=fillna,
            recover_pressure_from_mslp=True,
            drop_mslp=True,
        )
        df, _ = remove_irregular_stations(
            df, moving=rm_moving_stations, duplicates=rm_duplicate_stations
        )
        if selected_stations is not None:
            # Leave only selected stations
            df = df[df["id"].isin(selected_stations)]
            if len(df.index) == 0:
                raise DataNotFoundError
    except DataNotFoundError:
        logger.warning("No stations for DTG=%s", dtg)
        return

    if outdir_csv is not None:
        df["_f_rpath"] = obs_timestamp2csv_rpath(df["time_utc"])
        for f_rpath, file_df in df.groupby("_f_rpath", sort=False):
            fpath = Path(outdir_csv) / f_rpath
            file_df = file_df.drop("_f_rpath", axis=1).sort_values(
                by=["time_utc"]
            )
            save_df_as_netatmo_csv(file_df, fpath, overwrite=True)
        df = df.drop("_f_rpath", axis=1)

    if outdir_obsoul is not None:
        fpath = Path(outdir_obsoul) / "OBSOUL{}".format(
            dtg.strftime("%Y%m%d%H")
        )
        logger.debug(
            "%sSaving OBSOUL: DTG=%s, %d obs%s, file %s",
            logcolor.cyan,
            dtg,
            len(df.index),
            logcolor.reset,
            fpath,
        )
        save_df_as_obsoul(df, fpath, export_params=obsoul_export_params)


def netatmoqc_input2output(
    dtgs,
    netatmo_data_rootdir,
    selected_stations=None,
    dropna=True,
    fillna=None,
    rm_duplicate_stations=True,
    rm_moving_stations=True,
    outdir=None,
    loglevel="INFO",
    use_mpi=False,
    save_csv=True,
    save_obsoul=False,
    obsoul_export_params=None,
):
    """Save NetAtmoQC output data in CSV or OBSOUL formats.

    If a list of selected stations is passed in the "selected_stations", then
    only these are kept.

    Used in the "select" and "csv2obsoul" commands.

    Args:
        dtgs (netatmoqc.dtgs.DtgContainer): Assimilation DTGs for the data.
        netatmo_data_rootdir (pathlib.Path): Path to the top-level directory
            where the netatmo input csv files are stored.
        selected_stations (list): Selected station IDs (Default value = None).
        dropna (bool): Drop NaN values before saving (Default value = True).
        fillna (dict): Dictionary having data parameters as keys and the fill
            values as dict values. Passed to read_netatmo_csv
            (Default value = None).
        rm_duplicate_stations (bool): (Default value = True)
        rm_moving_stations (bool): (Default value = True)
        outdir (pathlib.Path): Dir where out files will be put.
            (Default value = ".")
        loglevel (str): Level for logging (Default value = "INFO").
        use_mpi (bool): Whether to use MPI or not (Default value = False).
        save_csv (bool): Whether to save csv files or not.
            (Default value = True).
        save_obsoul (bool): Whether to save obsoul files or not.
            (Default value = False)
        obsoul_export_params (list): List of data params to be exported in the
            final obsoul file. Passed to save_df_as_obsoul.
            (Default value = None).

    """
    if outdir is None:
        # Don't perform function calls in argument defaults
        outdir = Path()

    logger.info(
        "%sSaving selected observations...%s", logcolor.cyan, logcolor.reset,
    )

    outdir_csv = None
    outdir_obsoul = None
    if save_csv:
        outdir_csv = Path(outdir) / "csv_files"
        logger.info(
            "%s> CSV outdir:%s %s", logcolor.cyan, logcolor.reset, outdir_csv,
        )
    if save_obsoul:
        outdir_obsoul = Path(outdir) / "obsoul_files"
        logger.info(
            "%s> OBSOUL outdir:%s %s",
            logcolor.cyan,
            logcolor.reset,
            outdir_obsoul,
        )
        if obsoul_export_params is not None:
            logger.info(
                "    * OBSOUL files will contain: %s",
                ", ".join(obsoul_export_params),
            )

    in2out_fdtg = partial(
        _input2output_single_dtg,
        netatmo_data_rootdir=netatmo_data_rootdir,
        selected_stations=selected_stations,
        dropna=dropna,
        fillna=fillna,
        rm_duplicate_stations=rm_duplicate_stations,
        rm_moving_stations=rm_moving_stations,
        loglevel=loglevel,
        outdir_csv=outdir_csv,
        outdir_obsoul=outdir_obsoul,
        obsoul_export_params=obsoul_export_params,
    )

    if use_mpi:
        logger.info("Parallelising tasks over DTGs using MPI")
        mpi_parallel(in2out_fdtg, dtgs)
    else:
        n_procs = int(
            os.getenv("NETATMOQC_MAX_PYTHON_PROCS", psutil.cpu_count())
        )
        if n_procs > 1:
            logger.info("Parallelising tasks over DTGs within a single host")
        # Using the "loky" backend helps avoiding oversubscription
        # of cpus in child processes
        with parallel_backend("loky"):
            Parallel(n_jobs=n_procs)(delayed(in2out_fdtg)(dtg) for dtg in dtgs)
